from pathlib import Path
import shlex
import subprocess

from helix.configuration import MANAGER_CFG
from helix.data.file_type import FileType
from helix.utility.external import External


class Samtools:
    def __init__(
        self,
        external: External = External(),
        config=MANAGER_CFG.EXTERNAL,
        repo=MANAGER_CFG.REPOSITORY,
    ) -> None:
        self._external = external
        self._config = config
        self._repository_config = repo

    def fasta_index(self, input: Path, output: Path = None, regions=None, io=None):
        faidx_opt = ["faidx", input]
        if regions is not None:
            faidx_opt.append(regions)
        if output is not None:
            faidx_opt.extend(["-o", output])
        return self._external.samtools(faidx_opt, stdout=subprocess.PIPE, io=io)

    def consensus(self, input: Path, output: Path, stdin=None, io=None):
        consensus_opt = f'consensus "{input!s}" -o "{output!s}"'
        consensus_opt = shlex.split(consensus_opt)
        return self._external.samtools(consensus_opt, stdin=stdin, io=io)

    def header(self, file: Path, *args):
        return self._external.samtools(
            ["view", "-H", "--no-PG", *args, file], wait=True, text=True
        ).split("\n")

    def make_dictionary(self, path: Path, output: Path = None):
        if output is None:
            output = Path(path.parent, path.name + ".dict")
        self._external.samtools(["dict", str(path), "-o", str(output)], wait=True)

    def index(self, path: Path, wait=True, io=None):
        self._external.samtools(
            ["index", "-@", self._config.threads, "-b", str(path)], wait=wait, io=io
        )

    def index_stats(self, path: Path, wait=True, io=None):
        self._external.samtools(
            ["idxstats", path, "-@", self._config.threads], wait=wait, io=io
        )

    def sort(
        self, file: Path, output: Path = None, format: FileType = FileType.SAM, io=None
    ):
        options = ["sort", "-T", self._repository_config.temporary, "-n"]
        options.extend(["-@", self._config.threads])
        if output is not None:
            options.extend(["-o", str(output)])
        if format == FileType.SAM:
            options.extend(["-O", "sam"])
        elif format == FileType.BAM:
            options.extend(["-O", "bam"])
        else:
            raise RuntimeError("Unsupported file type")

        options.append(str(file))
        return self._external.samtools(options, io=io)

    def view(
        self,
        file: Path,
        target_format=FileType.SAM,
        output: Path = None,
        regions: str = None,
        subsample: float = None,
        cram_reference: Path = None,
        header=False,
        uncompressed=False,
        io=None,
    ):
        options = ["view", "--no-PG"]
        if target_format == FileType.BAM:
            options.append("-b")
        elif target_format == FileType.CRAM:
            options.extend(["-C", "-T", str(cram_reference)])
        elif target_format == FileType.SAM:
            pass
        else:
            raise ValueError("Cannot view in format %s" % (target_format.name))

        if header:
            options.append("-h")
        if uncompressed:
            options.append("-u")

        options.append(str(file))
        if output is not None:
            options.append("-o")
            options.append(output)

        if regions is not None:
            options.append(regions)

        if subsample is not None:
            options.extend(["-s", str(subsample)])

        options.extend(["-@", self._config.threads])

        return self._external.samtools(options, wait=True, io=io, text=True).split("\n")



import typing

from helix.data.chromosome_name_type import ChromosomeNameType


# Deprecated
class SequenceOrderer:
    def __init__(
        self,
        sequences=typing.List[str],
        target: ChromosomeNameType = ChromosomeNameType.Number,
    ) -> None:
        self.target = target
        # Map a canonic name with its index in the original sequences.
        self.sequence_map: dict[str, int] = {
            SequenceOrderer.canonicalize(x[1]): x[0] for x in enumerate(sequences)
        }
        self.ordered_sequences = self._get_ordered()

    def __iter__(self):
        """Generate an ordered list of sequences

        Yields:
            Tuple[int, str]: Tuple containing the index in the
                original sequence its converted name.

        """
        for sequence in self.ordered_sequences:
            yield self.sequence_map[sequence], SequenceOrderer.convert(
                sequence, self.target
            )

    def _get_ordered(self):
        autosome = self._get_autosome()
        sexual = self._get_sexual()
        mitochondrial = self._get_mitochondrial()
        others = self._get_others(autosome, sexual, mitochondrial)
        merged = [*autosome, *sexual, *mitochondrial, *others]
        return merged

    def _get_autosome(self):
        autosome = [x for x in self.sequence_map if x.isnumeric()]
        autosome.sort(key=lambda x: int(x))
        return autosome

    def _get_mitochondrial(self):
        return [x for x in self.sequence_map if x == "m"]

    def _get_sexual(self):
        sexual = []
        if "x" in self.sequence_map:
            sexual.append("x")
        if "y" in self.sequence_map:
            sexual.append("y")
        return sexual

    def _get_others(self, autosome, sexual, mitochondrial):
        others = []
        for sequence in self.sequence_map.keys():
            is_autosome = sequence in autosome
            is_sexual = sequence in sexual
            is_mitochondrial = sequence in mitochondrial
            if not is_autosome and not is_sexual and not is_mitochondrial:
                others.append(sequence)
        others.sort()
        return others

    def canonicalize(sequence_name: str) -> str:
        return SequenceOrderer.convert(sequence_name, ChromosomeNameType.Number)

    def convert(input: str, target: ChromosomeNameType):
        normalized = input.lower()
        if normalized.startswith("chr"):
            normalized = normalized.replace("chr", "", 1)
        if normalized.startswith("mt"):
            normalized = normalized.replace("mt", "m", 1)
        if target == ChromosomeNameType.Chr:
            if normalized.isnumeric() or normalized == "m":
                return "chr" + normalized
            return normalized
        elif target == ChromosomeNameType.Number:
            return normalized
        elif target == ChromosomeNameType.GenBank:
            raise NotImplementedError(
                "Converting to Accession is not supported at the moment."
            )
        raise ValueError(f"Converting to unrecognized target format: {target.name}")



from test.test_fasta import MockPath

import pytest

from helix.configuration import RepositoryConfig
from helix.microarray.microarray_converter import MicroarrayConverter

# def test_every_target_has_formatter():
#     assert all(x in TARGET_FORMATTER_MAP for x in MicroarrayConverterTarget)


def test_inexistent_head_folder_raise():
    config = RepositoryConfig()
    config.metadata = MockPath(exists=False)
    microarray_templates = MockPath(exists=False, name="microarray_templates")
    microarray_templates.joinpath = lambda x: MockPath(
        [], False if x == "body" else True, x
    )
    config.metadata.joinpath = lambda _: microarray_templates
    with pytest.raises(FileNotFoundError) as e:
        MicroarrayConverter(config)
    assert "template body"


def test_inexistent_head_folder_raise():
    config = RepositoryConfig()
    config.metadata = MockPath(exists=False)
    microarray_templates = MockPath(exists=False, name="microarray_templates")
    microarray_templates.joinpath = lambda x: MockPath(
        [], False if x == "head" else True, x
    )
    config.metadata.joinpath = lambda _: microarray_templates
    with pytest.raises(FileNotFoundError) as e:
        MicroarrayConverter(config)
    assert "template head"


def test_inexistent_head_folder_raise():
    config = RepositoryConfig()
    config.metadata = MockPath(exists=False)
    microarray_templates = MockPath(exists=False, name="microarray_templates")
    microarray_templates.joinpath = lambda x: MockPath(
        [], False if x == "body" else True, x
    )
    config.metadata.joinpath = lambda _: microarray_templates
    with pytest.raises(FileNotFoundError) as e:
        MicroarrayConverter(config)
    assert "template body"


# @patch('pathlib.Path.exists')
# @patch('pathlib.Path.open')
# @patch('gzip.open')
# @patch('shutil.copy')
# def test_missing_path(mock_copy, mock_gzip_open, mock_open, mock_exists):
#     output_lines = []
#     template_body = MockFile(["rs1\t1\t1"])
#     input_file = MockFile(["rs1\t1\t1\tAA"])
#     output_file = MockFile([])
#     output_file.write = lambda x: output_lines.append(x)

#     mock_open.side_effect = [input_file, output_file]
#     mock_exists.side_effect = [True, True, True, True]
#     mock_gzip_open.side_effect = [template_body]

#     config = RepositoryConfig()
#     config.metadata = Path("blah")
#     sut = MicroarrayConverter(config)
#     sut.convert(Path("foo"), MicroarrayConverterTarget.Ancestry_v1)
#     pass



import webbrowser

from helix.adapters.alignment_map_file_info_adapter import AlignmentMapFileInfoAdapter
from helix.adapters.alignment_stats_adapter import AlignmentStatsAdapter
from helix.adapters.header_adapter import HeaderAdapter
from helix.adapters.index_stats_adapter import IndexStatsAdapter
from helix.alignment_map.alignment_map_file import AlignmentMapFile
from helix.configuration import MANAGER_CFG
from helix.renderers.html_aligned_file_report import HTMLAlignedFileReport

if __name__ == "__main__":
    """This script will generate a report for
    the last file that was opened by Helix."""
    input = MANAGER_CFG.GENERAL.last_path
    if input is None:
        exit()
    if not input.exists():
        exit()
    if not input.is_file():
        exit()

    file = AlignmentMapFile(input)

    file_info_adapted = AlignmentMapFileInfoAdapter.adapt(file.file_info)
    index_stat_adapted = IndexStatsAdapter.adapt(file.file_info.index_stats)
    alignment_stats_adapted = AlignmentStatsAdapter.adapt(
        file.file_info.alignment_stats
    )
    header_adapted = HeaderAdapter.adapt(file.header)

    html_report = HTMLAlignedFileReport()
    html_page = html_report.make(
        file_info_adapted,
        index_stat_adapted,
        alignment_stats_adapted,
        header_adapted.sequences,
        header_adapted.read_groups,
        header_adapted.programs,
        header_adapted.comments,
    )

    extensions = "".join(MANAGER_CFG.GENERAL.last_path.suffixes[:-1])
    name = MANAGER_CFG.GENERAL.last_path.stem + extensions + ".html"
    target = input.with_name(name)

    with target.open("wt") as f:
        f.write(html_page)

    webbrowser.open(target)



import csv
import logging
import math
import statistics
import typing
from pathlib import Path

from helix.fasta.fasta_letter_counter import FASTALetterCounter
from helix.fasta.letter_run_buckets import LetterRunBuckets
from helix.fasta.letter_run_collection import LetterRunCollection


class FASTAStatsFiles:
    """Manage the creation of statistics files for Ns."""

    def __init__(
        self,
        fasta_file: FASTALetterCounter,
        long_run_threshold: int = 300,
        buckets_number: int = 1000,
    ) -> None:
        self._long_run_threshold = long_run_threshold
        self._buckets_number = buckets_number
        self._fasta_file = fasta_file

    def save_nbin(self, sequences: typing.List[LetterRunCollection]):
        lines = [
            "#WGS Extract runs of N: BIN definition file\n",
            f"#Processing Ref Model: {self._fasta_file.model_name} with >{self._long_run_threshold}bp of N runs\n",
            "#SN\tBinID\tStart\tSize\n",
        ]
        for sequence in sequences:
            for index, run in enumerate(
                sequence.filter(lambda x: x.length > self._long_run_threshold)
            ):
                row = f"{sequence.name}\t{index+1}\t{run.start:,}\t{run.length:,}\n"
                lines.append(row)
        return lines

    def load_bed(self):
        pass

    def save_bed(self, sequences: typing.List[LetterRunCollection]):
        lines = [
            "#WGS Extract runs of N: BED file of bin definitions\n",
            f"#Processing Ref Model: {self._fasta_file.model_name} with >{self._long_run_threshold}bp of N runs\n",
            "#SN\tStart\tStop\n",
        ]

        for sequence in sequences:
            for run in sequence.filter(lambda x: x.length > self._long_run_threshold):
                row = f"{sequence.name}\t{run.start}\t{run.start+run.length}\n"
                lines.append(row)
        return lines

    def load_nbuc(self):
        entries = list()
        with open(self._fasta_file.genome.nbuc, "r") as f:
            reader = csv.DictReader(f, delimiter="\t")
            for row in reader:
                entries.append
                sequence = LetterRunCollection(row[0], row[1])

                n_count = row[0]
                NumNreg = row[0]
                NregSizeMean = row[0]
                NregSizeStdDev = row[0]
                SmlNreg = row[0]
                BuckSize = row[0]

    def save_nbuc(self, sequences: typing.List[LetterRunCollection]):
        lines = [
            "#WGS Extract generated Sequence of N summary file\n",
            f"#Model {self._fasta_file.model_name} with >{self._long_run_threshold}bp Sequence of N and {self._buckets_number} buckets per sequence\n",
            "#Seq\tNumBP\tNumNs\tNumNreg\tNregSizeMean\tNregSizeStdDev\tSmlNreg\tBuckSize\tBucket Sparse List (bp start, ln value) when nonzero\n",
        ]
        for sequence in sequences:
            bucket_length = int(math.floor(sequence.length / self._buckets_number))

            long_runs = sequence.filter(lambda x: x.length > self._long_run_threshold)
            short_runs = sequence.filter(lambda x: x.length <= self._long_run_threshold)

            long_run_lengths = [x.length for x in long_runs]
            short_run_lengths = [x.length for x in short_runs]

            long_runs_total = sum(long_run_lengths)
            short_runs_total = sum(short_run_lengths)

            long_run_avg = 0
            long_run_stdev = 0

            if len(long_run_lengths) > 1:
                long_run_avg = int(statistics.mean(long_run_lengths))
                long_run_stdev = int(statistics.stdev(long_run_lengths))

            row = f"{sequence.name}\t{sequence.length}\t{long_runs_total}\t{len(long_run_lengths)}\t{long_run_avg}\t{long_run_stdev}\t{short_runs_total}\t{bucket_length}"

            buckets = LetterRunBuckets(
                sequence, self._buckets_number, self._long_run_threshold
            )
            for index, bucket_count in buckets.buckets.items():
                val = round(math.log(bucket_count) if bucket_count > 1 else 0)
                start = index * bucket_length

                if val > 0:
                    row += f"\t{start}\t{val}"
            row += "\n"
            lines.append(row)
        return lines

    def _generate_file(self, path: Path, lines: typing.List[str]):
        with path.open("wt", encoding="utf8") as f:
            f.writelines(lines)

    def _generate_files(
        self, sequences: typing.Dict[str, typing.List[LetterRunCollection]]
    ):
        self._generate_file(self._fasta_file.genome.nbuc, self.save_nbuc(sequences))
        self._generate_file(self._fasta_file.genome.bed, self.save_bed(sequences))
        self._generate_file(self._fasta_file.genome.nbin, self.save_nbin(sequences))

    def _load_files(self):
        pass

    def generate_stats(self):
        logging.info(f"{self._fasta_file.genome!s}: Counting Ns.")
        sequences = self._fasta_file.count_letters("N")
        logging.info(f"{self._fasta_file.genome!s}: Finished counting Ns.")
        self._generate_files(sequences)



import json
import re

from helix.configuration import MANAGER_CFG


class Sequencer:
    def __init__(self, name, sequencer_regex, coord_regex, coord_order, url) -> None:
        self.name = name
        self.sequencer_regex = re.compile(sequencer_regex)
        self.coord_regex = re.compile(coord_regex)
        self.coord_order = coord_order
        self.url = url


class Sequencers:
    def __init__(
        self,
        config=MANAGER_CFG.REPOSITORY,
    ) -> None:
        self._config = config
        self.data = self._load()

    def _load(self):
        content = None
        sequencer = self._config.metadata.joinpath("sequencers.json")
        with sequencer.open("rt") as p:
            content = json.load(p)
        return [Sequencer(**x) for x in content]

    def determine_sequencer(self, template_query):
        for sequencer in self.data:
            if sequencer.sequencer_regex.match(template_query):
                return sequencer.name
        return f"Unknown ({template_query})"



from test.utility import MockFile, MockPath

import pytest

from helix.fasta.fasta_letter_counter import FASTALetterCounter


class MockGenome:
    def __init__(self, fasta, dict) -> None:
        self.fasta = fasta
        self.dict = dict


def gzip_open(path, mode):
    return MockFile(path.lines)


def test_no_dictionary():
    fa_lines = [
        ">1 irrelevant\n",
        f"{'N'*5}\n",
        f"{'N'*5}\n",
        ">2 irrelevant\n",
    ]
    dict_lines = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(fa_lines), MockPath(None, False))
        sut = FASTALetterCounter(genome)
    assert "Unable to find dictionary" in str(e.value)


def test_sequence_not_in_dictionary():
    fa_lines = [
        ">1 irrelevant\n",
        f"{'N'*5}\n",
        f"{'N'*5}\n",
        ">2 irrelevant\n",
    ]
    dict_lines = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(ValueError) as e:
        genome = MockGenome(MockPath(fa_lines), MockPath(dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "not present in dictionary" in str(e.value)


def test_fastq():
    fa_lines = [
        ">1 irrelevant\n",
        f"{'N'*5}\n",
        f"{'N'*5}\n",
        "+2 irrelevant\n",
    ]
    dict_lines = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(fa_lines), MockPath(dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "Expected a FASTA" in str(e.value)


def test_duplicate_sequence():
    fa_lines = [
        ">1 irrelevant\n",
        f"{'N'*5}\n",
        f"{'N'*5}\n",
        ">1 irrelevant\n",
    ]
    dict_lines = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(fa_lines), MockPath(dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "duplicated sequence" in str(e.value)


def test_only_comments():
    fa_lines = ["#Hello\n", f"#Foo\n"]
    dict_lines = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)

    with pytest.raises(RuntimeError) as e:
        genome = MockGenome(MockPath(fa_lines), MockPath(dict_lines))
        FASTALetterCounter(genome).count_letters()
    assert "no sequences" in str(e.value)



import gzip
from pathlib import Path


class RawEntry:
    def __init__(self, id=None, chromosome=None, position=None, result=None) -> None:
        self.id: str = id
        self.chromosome: str = chromosome
        self.position: int = position
        self.result: str = result

    @property
    def template_entry(self):
        return f"{self.id}\t{self.chromosome}\t{self.position}\n"

    def __str__(self) -> str:
        return f"{self.id},{self.chromosome},{self.position},{self.result}"

    def __repr__(self) -> str:
        return self.__str__()


class RawFile:
    def __init__(self, path: Path) -> None:
        self.path = path
        if self.path.suffix == ".gz":
            self.open_file = lambda: gzip.open(self.path, "rt")
        else:
            self.open_file = lambda: self.path.open("rt")

    def __iter__(self):
        with self.open_file() as file:
            for line in file:
                if "#" in line:
                    continue
                parts = line.split("\t")
                id = parts[0].strip()
                chromosome = parts[1].strip()
                position = int(parts[2])
                result = None
                if len(parts) > 3:
                    result = parts[3].strip()
                yield RawEntry(id, chromosome, position, result)



from enum import Flag, auto
import uuid

from helix.configuration import MANAGER_CFG
from helix.data.sequence import Sequence
from helix.naming.converter import Converter


class RegionType(Flag):
    Y = auto()
    MT = auto()
    WES = auto()


class Regions:
    def __init__(self, config=MANAGER_CFG.REPOSITORY, converter=Converter()) -> None:
        self._templates = config.metadata.joinpath("bed_templates")
        self._temporary = config.temporary
        self._converter = converter

    def get_path(self, build: str, type: RegionType, sequences: list[Sequence]):
        name_map = {x.canonic_name: x.name for x in sequences}

        suffix = ""
        if type & RegionType.Y and type & RegionType.WES:
            raise ValueError("Invalid combination: Y and WES")
        if type & RegionType.MT and type & RegionType.WES:
            raise ValueError("Invalid combination: MT and WES")

        if type & RegionType.Y:
            suffix += "_y"
        if type & RegionType.MT:
            suffix += "_mt"
        if type & RegionType.WES:
            suffix += "_wes"
        file_name = f"{build}{suffix}.csv"
        file = self._templates.joinpath(file_name)
        if not file.exists():
            raise FileNotFoundError(f"Unable to find BED template at: {file!s}")
        with file.open("r") as f:
            lines = f.readlines()
            rows = []
            for line in lines:
                line = line.split(",")
                sequence = line[0]
                begin = line[1]
                end = line[2].strip()
                if sequence in name_map:
                    sequence = name_map[sequence]
                rows.append(f"{sequence}\t{begin}\t{end}\n")
        temporary_file = self._temporary.joinpath(str(uuid.uuid4()))
        with temporary_file.open("w") as f:
            f.writelines(rows)
        return temporary_file

    def get_content(self, build: str, type: RegionType):
        file = self.get_path(build, type)
        return file.read_text("utf-8")



from enum import Enum, IntFlag, auto


class AlignmentMapFlag(IntFlag):
    """Flags contained in a SAM row.
    Reference: https://samtools.github.io/hts-specs/SAMv1.pdf, Page 7.
    """

    MULTIPLE_SEGMENTS = 0x1
    ALIGNED = 0x2
    UNMAPPED = 0x4
    NEXT_SEGMENT_UNMAPPED = 0x8
    SEQ_REVERSE_COMPLEMENTED = 0x10
    NEXT_SEQ_REVERSE_COMPLEMENTED = 0x20
    FIRST_SEGMENT = 0x40
    LAST_SEGMENT = 0x80
    SECONDARY_ALIGNMENT = 0x100
    NOT_PASSING = 0x200
    DUPLICATE = 0x400
    SUPPLEMENTARY_ALIGNMENT = 0x800


class AlignmentMapOptionalFieldType(Enum):
    UNKNOWN = auto()
    PRINTABLE = "A"
    SIGNED_INT = "i"
    SINGLE_FLOAT = "f"
    PRINTABLE_WITH_SPACES = "Z"
    HEX_ARRAY = "H"
    NUMERIC_ARRAY = "B"


class AlignmentMapOptionalFieldArrayType(Enum):
    UNKNOWN = auto()
    INT_8 = "c"
    UINT_8 = "C"
    INT_16 = "s"
    UINT_16 = "S"
    INT_32 = "i"
    UINT_32 = "I"
    FLOAT = "f"


class AlignmentMapOptionalField:
    def __init__(self, field: str) -> None:
        tag, type, value = field.split(":", 2)

        self.tag = tag
        self.type = self._type_to_enum(type)
        self.value = self._get_value(value)

    def _type_to_enum(self, type: str):
        if type in [x.value for x in AlignmentMapOptionalFieldType]:
            return AlignmentMapOptionalFieldType(type)
        return AlignmentMapOptionalFieldType.UNKNOWN

    def _get_value(self, value):
        if self.type == AlignmentMapOptionalFieldType.SIGNED_INT:
            return int(value)
        elif self.type == AlignmentMapOptionalFieldType.SINGLE_FLOAT:
            return float(value)
        elif self.type == AlignmentMapOptionalFieldType.HEX_ARRAY:
            return bytes.fromhex(value)
        elif self.type == AlignmentMapOptionalFieldType.NUMERIC_ARRAY:
            return self._process_array(value)
        return value

    def _process_array(self, value):
        array_type = AlignmentMapOptionalFieldArrayType.UNKNOWN
        if value[0] in [x.value for x in AlignmentMapOptionalFieldArrayType]:
            array_type = AlignmentMapOptionalFieldArrayType(value[0])
        if array_type == AlignmentMapOptionalFieldArrayType.UNKNOWN:
            return value
        values = value.split(",")
        if len(values) == 1:
            return []
        values = values[1::]
        if array_type == AlignmentMapOptionalFieldArrayType.FLOAT:
            return [float(x) for x in values]
        return [int(x) for x in values]

    def __str__(self) -> str:
        return f"{self.tag}: {self.value}"

    def __repr__(self) -> str:
        return self.__str__()


class AlignmentMapRow:
    """Parser for a SAM row.
    Reference: https://samtools.github.io/hts-specs/SAMv1.pdf, Page 7+
    """

    def __init__(self, row: str) -> None:
        self.__row: str = row
        self.__values: list[str] = None
        self.__query_template_name: str = None
        self.__values = None
        self.__flag: AlignmentMapFlag = None
        self.__reference_sequence_name: str = None
        self.__position: int = None
        self.__mapping_quality: int = None
        self.__cigar: str = None
        self.__mate_sequence_name: str = None
        self.__mate_position: int = None
        self.__template_length: int = None
        self.__sequence: str = None
        self.__quality: list[int] = None
        self.__optional: list[AlignmentMapOptionalField] = None

    @property
    def values(self):
        if self.__values is None:
            self.__values = self.__row.split()
        return self.__values

    @property
    def query_template_name(self):
        """ID for a specific read, * if unavailable.
        Usually this ID is unique except few cases.
        See the reference for more details."""
        if self.__query_template_name is None:
            self.__query_template_name = self.values[0]
        return self.__query_template_name

    @property
    def flag(self):
        """Combination of bitwise flags for this read"""
        if self.__flag is None:
            self.__flag = AlignmentMapFlag(int(self.values[1]))
        return self.__flag

    @property
    def reference_sequence_name(self):
        """Name of the sequence. Either a value contained in the
        header or * if unmapped."""
        if self.__reference_sequence_name is None:
            self.__reference_sequence_name = self.values[2]
        return self.__reference_sequence_name

    @property
    def position(self):
        """1-based position of the 1st CIGAR op that "consumes"
        a reference base (see specific for CIGAR at page 8).
        '*' if unmapped."""
        if self.__position is None:
            self.__position = int(self.values[3])
        return self.__position

    @property
    def mapping_quality(self):
        """Mapping quality, equal to -10log_10 Pr{mapping position is wrong}"""
        if self.__mapping_quality is None:
            self.__mapping_quality = int(self.values[4])
        return self.__mapping_quality

    @property
    def cigar(self):
        """CIGAR string"""
        if self.__cigar is None:
            self.__cigar = self.values[5]
        return self.__cigar

    @property
    def mate_sequence_name(self):
        """Reference sequence name of the primary alignment of the mate read.
        * if unavailable, = if identical to current reference sequence name.
        Should match reference sequence name of the mate if it has one primary
        mapping."""
        if self.__mate_sequence_name is None:
            self.__mate_sequence_name = self.values[6]
        return self.__mate_sequence_name

    @property
    def mate_position(self):
        """1-based position of the primary alignment of the mate read in the template.
        0 if unavailable. Should match the position of the mate read."""
        if self.__mate_position is None:
            self.__mate_position = int(self.values[7])
        return self.__mate_position

    @property
    def template_length(self):
        if self.__template_length is None:
            self.__template_length = int(self.values[8])
        return self.__template_length

    @property
    def sequence(self):
        """The sequence of bases for this read"""
        if self.__sequence is None:
            self.__sequence = self.values[9]
        return self.__sequence

    @property
    def quality(self):
        """Defined as -10 log_10 Pr{base is wrong}. One number for each base."""
        if self.__quality is None:
            self.__quality = [ord(x) - 33 for x in self.values[10]]
        return self.__quality

    @property
    def optional(self):
        """Optional fields"""
        if self.__optional is None:
            if len(self.values) <= 10:
                return []
            self.__optional = [AlignmentMapOptionalField(x) for x in self.values[11::]]
        return self.__optional

    def __str__(self) -> str:
        return (
            f"{self.reference_sequence_name}:{self.position}; Q: {self.mapping_quality}"
        )

    def __repr__(self) -> str:
        return self.__str__()



from PySide6.QtGui import QFont
from PySide6.QtWidgets import (
    QCheckBox,
    QGridLayout,
    QGroupBox,
    QLabel,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from helix.microarray.microarray_converter import MicroarrayConverterTarget


class MicroarraySelection(QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.setObjectName(MicroarraySelection.__name__)
        self.main_group_box = QGroupBox("Select microarray format", parent)
        self.inner_layout = QVBoxLayout(self.main_group_box)
        self.main_layout = QVBoxLayout()

        self.main_group_box.setObjectName("mainGroupBox")
        self.inner_layout.setObjectName("innerLayout")

        self.combined_checkboxes = None
        self.genealogy_checkboxes = None
        self.health_checkboxes = None
        self.ancient_dna_checkboxes = None

        self.combined_group = self._create_combined_group_box(self.main_group_box)
        self.genealogy_group = self._create_genealogy_group_box(self.main_group_box)
        self.health_group = self._create_health_group_box(self.main_group_box)
        self.ancient_group = self._create_ancient_dna_group_box(self.main_group_box)

        self.checkboxes = [
            *self.combined_checkboxes,
            *self.genealogy_checkboxes,
            *self.health_checkboxes,
            *self.ancient_dna_checkboxes,
        ]

        self.inner_layout.addWidget(self.combined_group)
        self.inner_layout.addWidget(self.genealogy_group)
        self.inner_layout.addWidget(self.health_group)
        self.inner_layout.addWidget(self.ancient_group)

        self.main_layout.addWidget(self.main_group_box)
        self.setLayout(self.main_layout)
        self.hide()

    def _create_combined_group_box(self, parent):
        boldFont = QFont()
        boldFont.setBold(True)

        combined_group = QGroupBox("Combined Kits", parent)
        layout = QGridLayout(combined_group)

        combinedWGSELabel = QLabel("WGSE", combined_group)
        combined23AndMeLabel = QLabel("23AndMe", combined_group)
        combinedReichLabLabel = QLabel("ReichLab", combined_group)

        self.combined_checkboxes = [
            QCheckBox("All", combined_group),
            QCheckBox("v3 *", combined_group),
            QCheckBox("v5", combined_group),
            QCheckBox("v3 + v5 *", combined_group),
            QCheckBox("API SNPs", combined_group),
            QCheckBox("1240K+HumanOrigins", combined_group),
        ]

        combinedWGSELabel.setFont(boldFont)
        combined23AndMeLabel.setFont(boldFont)
        combinedReichLabLabel.setFont(boldFont)

        combined_group.setObjectName("combinedKitsGroupBox")
        layout.setObjectName("combinedLayout")

        combinedWGSELabel.setObjectName("combinedWGSELabel")
        combined23AndMeLabel.setObjectName("combined23AndMeLabel")
        combinedReichLabLabel.setObjectName("combinedReichLabLabel")

        self.combined_checkboxes[0].setObjectName(MicroarrayConverterTarget.All.name)
        self.combined_checkboxes[1].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_v3.name
        )
        self.combined_checkboxes[2].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_v5.name
        )
        self.combined_checkboxes[3].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_v35.name
        )
        self.combined_checkboxes[4].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_SNPs_API.name
        )
        self.combined_checkboxes[5].setObjectName(
            MicroarrayConverterTarget.ReichLab_HumanOrigins_v1.name
        )

        layout.addWidget(combinedWGSELabel, 0, 0, 1, 1)
        layout.addWidget(combined23AndMeLabel, 1, 0, 1, 1)
        layout.addWidget(combinedReichLabLabel, 2, 0, 1, 1)

        layout.addWidget(self.combined_checkboxes[0], 0, 1, 1, 1)
        layout.addWidget(self.combined_checkboxes[1], 0, 2, 1, 1)
        layout.addWidget(self.combined_checkboxes[2], 0, 3, 1, 1)
        layout.addWidget(self.combined_checkboxes[3], 1, 1, 1, 1)
        layout.addWidget(self.combined_checkboxes[4], 1, 2, 1, 1)
        layout.addWidget(self.combined_checkboxes[5], 2, 1, 1, 1)
        return combined_group

    def _create_genealogy_group_box(self, parent):
        boldFont = QFont()
        boldFont.setBold(True)

        genealogy_group = QGroupBox("Genealogy", parent)
        layout = QGridLayout(genealogy_group)

        genealogyAncestryDNALabel = QLabel("Ancestry", genealogy_group)
        genealogy23AndMeLabel = QLabel("23AndMe", genealogy_group)
        genealogyFamilyTreeDNALabel = QLabel("FamilyTreeDNA", genealogy_group)
        genealogyLivingDNALabel = QLabel("Living DNA", genealogy_group)
        genealogyMyHeritageLabel = QLabel("MyHeritage", genealogy_group)

        self.genealogy_checkboxes = [
            QCheckBox("v3", genealogy_group),
            QCheckBox("v4", genealogy_group),
            QCheckBox("v5", genealogy_group),
            QCheckBox("v1", genealogy_group),
            QCheckBox("v2", genealogy_group),
            QCheckBox("v1", genealogy_group),
            QCheckBox("v2", genealogy_group),
            QCheckBox("v3", genealogy_group),
            QCheckBox("v1", genealogy_group),
            QCheckBox("v2", genealogy_group),
            QCheckBox("v1", genealogy_group),
            QCheckBox("v2", genealogy_group),
        ]

        genealogy_group.setObjectName("genealogyGroupBox")
        layout.setObjectName("genealogyLayout")
        genealogy23AndMeLabel.setObjectName("genealogy23AndMeLabel")
        genealogyAncestryDNALabel.setObjectName("genealogyAncestryDNA")
        genealogyFamilyTreeDNALabel.setObjectName("genealogyFamilyTreeDNA")
        genealogyLivingDNALabel.setObjectName("genealogyLivingDNALabel")
        genealogyMyHeritageLabel.setObjectName("genealogyMyHeritageLabel")

        genealogy23AndMeLabel.setFont(boldFont)
        genealogyAncestryDNALabel.setFont(boldFont)
        genealogyFamilyTreeDNALabel.setFont(boldFont)
        genealogyLivingDNALabel.setFont(boldFont)
        genealogyMyHeritageLabel.setFont(boldFont)

        self.genealogy_checkboxes[0].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_v3.name
        )
        self.genealogy_checkboxes[1].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_v4.name
        )
        self.genealogy_checkboxes[2].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_v5.name
        )
        self.genealogy_checkboxes[3].setObjectName(
            MicroarrayConverterTarget.Ancestry_v1.name
        )
        self.genealogy_checkboxes[4].setObjectName(
            MicroarrayConverterTarget.Ancestry_v2.name
        )
        self.genealogy_checkboxes[5].setObjectName(
            MicroarrayConverterTarget.FTDNA_v1.name
        )
        self.genealogy_checkboxes[6].setObjectName(
            MicroarrayConverterTarget.FTDNA_v2.name
        )
        self.genealogy_checkboxes[7].setObjectName(
            MicroarrayConverterTarget.FTDNA_v3.name
        )
        self.genealogy_checkboxes[8].setObjectName(
            MicroarrayConverterTarget.LivingDNA_v1.name
        )
        self.genealogy_checkboxes[9].setObjectName(
            MicroarrayConverterTarget.LivingDNA_v2.name
        )
        self.genealogy_checkboxes[10].setObjectName(
            MicroarrayConverterTarget.MyHeritage_v1.name
        )
        self.genealogy_checkboxes[11].setObjectName(
            MicroarrayConverterTarget.MyHeritage_v2.name
        )

        layout.addWidget(genealogy23AndMeLabel, 0, 0, 1, 1)
        layout.addWidget(genealogyAncestryDNALabel, 2, 0, 1, 1)
        layout.addWidget(genealogyFamilyTreeDNALabel, 3, 0, 1, 1)
        layout.addWidget(genealogyLivingDNALabel, 4, 0, 1, 1)
        layout.addWidget(genealogyMyHeritageLabel, 5, 0, 1, 1)

        layout.addWidget(self.genealogy_checkboxes[0], 0, 1, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[1], 0, 2, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[2], 0, 3, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[3], 2, 1, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[4], 2, 2, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[5], 3, 1, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[6], 3, 2, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[7], 3, 3, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[8], 4, 1, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[9], 4, 2, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[10], 5, 1, 1, 1)
        layout.addWidget(self.genealogy_checkboxes[11], 5, 2, 1, 1)
        return genealogy_group

    def _create_ancient_dna_group_box(self, parent):
        boldFont = QFont()
        boldFont.setBold(True)

        ancient_group = QGroupBox("Ancient DNA", parent)
        layout = QGridLayout(ancient_group)
        reich_lab_label = QLabel("Reich Lab", ancient_group)

        reich_lab_label.setFont(boldFont)

        ancient_group.setObjectName("ancientDNAGroupBox")
        layout.setObjectName("ancientDNALayout")
        reich_lab_label.setObjectName("ancientDNAReichLabLabel")

        self.ancient_dna_checkboxes = [
            QCheckBox("AADR 1240K", ancient_group),
            QCheckBox("Human Origins v1", ancient_group),
        ]

        self.ancient_dna_checkboxes[0].setObjectName(
            MicroarrayConverterTarget.ReichLab_AADR.name
        )
        self.ancient_dna_checkboxes[1].setObjectName(
            MicroarrayConverterTarget.ReichLab_HumanOrigins_v1.name
        )

        spacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )

        layout.addWidget(reich_lab_label, 0, 0, 1, 1)
        layout.addWidget(self.ancient_dna_checkboxes[0], 0, 1, 1, 1)
        layout.addWidget(self.ancient_dna_checkboxes[1], 0, 2, 1, 1)
        layout.addItem(spacer, 0, 3, 1, 1)

        return ancient_group

    def _create_health_group_box(self, parent):
        health_group = QGroupBox("Health", parent)
        health_layout = QGridLayout(health_group)

        self.health_checkboxes = [
            QCheckBox("23AndMe", health_group),
            QCheckBox("TellMeGen", health_group),
            QCheckBox("MTHFR Genetics", health_group),
            QCheckBox("AncestryDNA", health_group),
            QCheckBox("SelfDecode", health_group),
        ]

        health_group.setObjectName("healthGroupBox")
        self.health_checkboxes[0].setObjectName(
            MicroarrayConverterTarget.TwentyThreeAndMe_v5.name
        )
        self.health_checkboxes[1].setObjectName(
            MicroarrayConverterTarget.TellMeGen.name
        )
        self.health_checkboxes[2].setObjectName(MicroarrayConverterTarget.MTHFR.name)
        self.health_checkboxes[3].setObjectName(
            MicroarrayConverterTarget.Ancestry_v2.name
        )
        self.health_checkboxes[4].setObjectName(
            MicroarrayConverterTarget.SelfDecode.name
        )

        spacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum
        )
        health_layout.setObjectName("healthLayout")
        health_layout.addItem(spacer, 0, 0, 2, 1)
        health_layout.addWidget(self.health_checkboxes[0], 0, 1, 1, 1)
        health_layout.addWidget(self.health_checkboxes[1], 0, 2, 1, 1)
        health_layout.addWidget(self.health_checkboxes[2], 0, 3, 1, 1)

        health_layout.addWidget(self.health_checkboxes[3], 1, 1, 1, 1)
        health_layout.addWidget(self.health_checkboxes[4], 1, 2, 1, 1)
        return health_group



import enum

from helix.data.alignment_stats import AlignmentStats
from helix.data.tabular_data import TabularData, TabularDataRow


class AlignmentStatsAdapter:
    def adapt(input_data: AlignmentStats):
        name_map = {
            "skipped_samples": "Skipped samples",
            "samples_count": "Samples count",
            "sequencer": "Sequencer",
            "duplicate": "Duplicates",
            "count_length": "Count length",
            "average_length": "Average length",
            "standard_dev_length": "Standard deviation length",
            "count_insert_size": "Count insert size",
            "average_insert_size": "Average insert size",
            "standard_dev_insert_size": "Standard deviation insert size",
            "count_quality": "Count quality",
            "average_quality": "Average quality",
            "standard_dev_quality": "Standard deviation quality",
            "read_type": "Read type",
        }

        rows = []
        for key, value in name_map.items():
            data = input_data.__dict__[key]
            if isinstance(data, enum.Enum):
                data = str(data.name)
            rows.append(TabularDataRow(value, [str(data)]))
        return TabularData(None, rows)



from pydantic import BaseModel
from helix.data.sorting import Sorting


class AlignmentMapHeaderMetadata(BaseModel):
    version: str = None
    sorted: Sorting = None
    grouping: str = None
    subsorting: str = None



from helix.alignment_map.alignment_map_header import AlignmentMapHeader
from helix.data.alignment_map.alignment_map_header_program import (
    AlignmentMapHeaderProgram,
)
from helix.data.alignment_map.alignment_map_header_read_group import (
    AlignmentMapHeaderReadGroup,
)
from helix.data.alignment_map.alignment_map_header_sequence import (
    AlignmentMapHeaderSequence,
)
from helix.data.tabular_data import TabularData, TabularDataRow


class HeaderSequenceAdapter:
    def adapt(input_data: list[AlignmentMapHeaderSequence]):
        return TabularData(
            [
                "Name",
                "Length",
                "MD5",
                "Alternate locus",
                "Alternative names",
                "Description",
                "Genome",
                "Molecule topology",
                "Species",
                "URI",
            ],
            [
                TabularDataRow(
                    None,
                    [
                        x.name,
                        str(x.length),
                        x.md5,
                        x.alternate_locus,
                        x.alternative_names,
                        x.description,
                        x.genome_assembly_identifier,
                        x.molecule_topology,
                        x.species,
                        x.uri,
                    ],
                )
                for x in input_data
            ],
        )


class HeaderProgramsAdapter:
    def adapt(input_data: list[AlignmentMapHeaderProgram]):
        return TabularData(
            ["ID", "Description", "Command line", "Previous", "Program version"],
            [
                TabularDataRow(
                    None,
                    [
                        x.id,
                        x.name,
                        x.description,
                        x.command_line,
                        x.previous,
                        x.program_version,
                    ],
                )
                for x in input_data
            ],
        )


class HeaderReadGroupAdapter:
    def adapt(input_data: list[AlignmentMapHeaderReadGroup]):
        return TabularData(
            [
                "ID",
                "Description",
                "Barcode",
                "Date",
                "Flow order",
                "Key sequence",
                "Library",
                "Platform",
                "Platform model",
                "Platform unit",
                "Predicted median insert size",
                "Sample",
                "Sequencing Center",
                "Programs",
            ],
            [
                TabularDataRow(
                    None,
                    [
                        x.id,
                        x.description,
                        x.barcode,
                        x.date,
                        x.flow_order,
                        x.key_sequence,
                        x.library,
                        x.platform,
                        x.platform_model,
                        x.platform_unit,
                        x.predicted_median_insert_size,
                        x.sample,
                        x.sequencing_center,
                        x.programs,
                    ],
                )
                for x in input_data
            ],
        )


class HeaderCommentsAdapter:
    def adapt(input_data: list[str]):
        return TabularData(["Text"], [TabularDataRow(None, [x for x in input_data])])


class AdaptedHeader:
    def __init__(self, sequences, read_groups, programs, comments) -> None:
        self.sequences: TabularData = sequences
        self.read_groups: TabularData = read_groups
        self.programs: TabularData = programs
        self.comments: TabularData = comments


class HeaderAdapter:
    def adapt(input_data: AlignmentMapHeader) -> AdaptedHeader:
        return AdaptedHeader(
            HeaderSequenceAdapter.adapt(input_data.sequences.values()),
            HeaderReadGroupAdapter.adapt(input_data.read_groups),
            HeaderProgramsAdapter.adapt(input_data.programs),
            HeaderCommentsAdapter.adapt(input_data.comments),
        )



import gzip
import hashlib
import logging
import shutil
import typing
import zipfile
from pathlib import Path

from helix.files.bgzip import BGzip, BgzipAction
from helix.reference.genome_metadata_loader import Genome
from helix.utility.external import External
from helix.files.file_type_checker import FileType, FileTypeChecker


class Decompressor:
    def __init__(
        self,
        type_checker: FileTypeChecker = FileTypeChecker(),
        external: External = External(),
        bgzip_compressor=BGzip(),
    ) -> None:
        self._bgzip_compressor = bgzip_compressor
        self._external = external
        self._type_checker = type_checker

        self._handlers: typing.Dict[FileType, typing.Callable[[Path, Path], None]] = {
            FileType.GZIP: Decompressor.razf_gzip,
            FileType.RAZF_GZIP: Decompressor.razf_gzip,
            FileType.ZIP: Decompressor.zip,
            FileType.SEVENZIP: Decompressor.sevenzip,
            FileType.BZIP: Decompressor.bzip,
        }

    def gz(self, input_file: Path, output_file: Path):
        # Not reliable with RAZF and currently not used.
        with gzip.open(str(input_file), "rb") as f_in:
            with open(output_file, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

    def sevenzip(self, input_file: Path, output_file: Path):
        raise NotImplementedError()

    def bzip(self, input_file: Path, output_file: Path):
        raise NotImplementedError()

    def zip(self, input_file: Path, output_file: Path):
        with zipfile.ZipFile(str(input_file), "r") as f:
            files = f.namelist()
            if len(files) > 1:
                raise RuntimeError(
                    "More than one file found inside the .zip, unable to proceed."
                )
            extracted = Path(f.extract(files[0], output_file.parent))
            if extracted != output_file:
                extracted.rename(output_file)

    def razf_gzip(self, input_file: Path, output_file: Path):
        self._bgzip_compressor.bgzip_wrapper(
            input_file, output_file, BgzipAction.Decompress
        )

    def calculate_md5_hash(self, filename: Path, chunk_size=4096):
        md5_hash = hashlib.md5()
        with filename.open("rb") as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                md5_hash.update(chunk)
        return md5_hash.hexdigest()

    def decompression_needed(self, genome: Genome, file: Path):
        if not file.exists():
            return True

        type = self._type_checker.get_type(file)
        # Shortcut in case we're dealing with a BGZIP file
        if type == FileType.BGZIP:
            if genome.bgzip_size is None:
                genome.bgzip_size = file.stat().st_size
                return False
            if file.stat().st_size == genome.bgzip_size:
                return False
        # Shortcut in case we're dealing with a FASTA file
        elif type == FileType.DECOMPRESSED:
            if genome.decompressed_size is None:
                return True
            if file.stat().st_size == genome.decompressed_size:
                return False
        return True

    def perform(self, genome: Genome, downloaded: Path = None):
        if not self.decompression_needed(genome, downloaded):
            return downloaded

        target = downloaded.with_suffix(".fa")
        if not self.decompression_needed(genome, target):
            return target

        if not downloaded.exists():
            raise FileNotFoundError(f"Unable to find file {downloaded.name}")

        type = self._type_checker.get_type(downloaded)
        if type not in self._handlers:
            raise RuntimeError(
                f"Trying to decompress a file with an unknown type: {downloaded.name}"
            )

        handler = self._handlers[type]
        logging.debug(
            f"Decompressing {genome.fasta.name}. {type.name} compression detected."
        )
        handler(self, downloaded, target)
        if not target.exists():
            raise RuntimeError(
                f"Error during the decompression of {str(genome)}. Decompressed file is not present."
            )

        if genome.decompressed_size is None:
            genome.decompressed_size = target.stat().st_size
        elif genome.decompressed_size != target.stat().st_size:
            raise RuntimeError(
                f"Error during the decompression of {str(genome)}. Size mismatch."
            )

        if genome.decompressed_md5 is None:
            genome.decompressed_md5 = self.calculate_md5_hash(target)
        elif genome.decompressed_md5 != self.calculate_md5_hash(target):
            raise RuntimeError(
                f"Error during the decompression of {str(genome)}. MD5 mismatch."
            )

        return target



import logging
import pickle
from pathlib import Path

from helix.alignment_map.alignment_map_header import AlignmentMapHeader
from helix.alignment_map.alignment_stats_calculator import AlignmentStatsCalculator
from helix.alignment_map.index_stats_calculator import (
    IndexStatsCalculator,
    SequenceStatistics,
)
from helix.configuration import MANAGER_CFG
from helix.data.alignment_map.alignment_map_file_info import AlignmentMapFileInfo
from helix.data.file_type import FileType
from helix.data.gender import Gender
from helix.data.mitochondrial_model_type import MitochondrialModelType
from helix.data.read_type import ReadType
from helix.data.sequence_type import SequenceType
from helix.data.sorting import Sorting
from helix.fasta.reference import Reference, ReferenceStatus
from helix.progress.base_progress_calculator import BaseProgressCalculator
from helix.reference.repository_manager import Repository
from helix.mtDNA.mt_dna import MtDNA
from helix.utility.samtools import Samtools

logger = logging.getLogger(__name__)


class AlignmentMapFile:
    SUPPORTED_FILES = {
        ".bam": FileType.BAM,
        ".sam": FileType.SAM,
        ".cram": FileType.CRAM,
    }

    def __init__(
        self,
        path: Path,
        samtools=Samtools(),
        repository: Repository = Repository(),
        mtdna: MtDNA = MtDNA(),
        config=MANAGER_CFG.EXTERNAL,
    ) -> None:
        if isinstance(path, str):
            path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Unable to find file {path.name}")
        if path.suffix.lower() not in AlignmentMapFile.SUPPORTED_FILES:
            raise RuntimeError(f"Unrecognized file extension: {path.name}")

        self.path: Path = path
        self.meta_file: Path = self.path.with_suffix(".pickle")
        self._repo = repository
        self._samtools = samtools
        self._mtdna = mtdna
        self._config = config
        self.header = self._load_header()
        self.file_info = self._initialize_file_info()

    def subset(self, percentage):
        if percentage < 0.0 or percentage > 100.0:
            raise RuntimeError(
                f"Percentage needs to be between 0 and 100 to produce a subset but was {percentage}"
            )

        output = self.path.with_stem(f"{self.path.stem}_{percentage}_percent")
        percentage /= 100
        reference = None
        if self.file_info.file_type == FileType.CRAM:
            reference = self.file_info.reference_genome.ready_reference
            if reference is None:
                raise FileNotFoundError(
                    "Reference genome was not found but is mandatory for CRAM files"
                )

        self._samtools.view(
            self.path, output, subsample=percentage, reference=reference
        )
        self._samtools.index(output)

        if not output.exists():
            raise RuntimeError(f"Unable to find the subset file at {output}")
        return AlignmentMapFile(output)

    def _load_header(self) -> AlignmentMapHeader:
        return AlignmentMapHeader(self._samtools.header(self.path))

    def _to_fastq(self, io):
        options_fq = ["fastq"]
        if self.file_info.alignment_stats.read_type == ReadType.Paired:
            name_1 = self.path.stem + "_R1.fastq.gz"
            name_2 = self.path.stem + "_R2.fastq.gz"
            name_1 = self.path.parent.joinpath(name_1)
            name_2 = self.path.parent.joinpath(name_2)
            options_fq.extend(["-1", str(name_1), "-2", str(name_2)])
        elif self.file_info.alignment_stats.read_type == ReadType.Single:
            name = self.path.stem + ".fastq.gz"
            options_fq.extend(["-0", str(name)])
        else:
            raise RuntimeError("Unknown read type")
        options_fq.append("-n")
        options_fq.extend(["-@", self._config.threads])

        view = self._samtools.view(
            self.path,
            target_format=FileType.BAM,
            cram_reference=self.file_info.reference_genome.ready_reference,
            header=True,
            uncompressed=True,
            io=io,
        )

        # sort = self._samtools.sort(self.path)
        # fastq = self._samtools.fastq(self.path, )

    def _to_fasta(self, regions=None, progress=None):
        input = self.path
        suffixes = self.path.suffixes.copy()
        suffixes[-1] = ".fasta"
        output = self.path.with_name(self.path.stem + "".join(suffixes))
        reference = str(self.file_info.reference_genome.ready_reference.fasta)

        io = None
        # TODO: get a percentage of input read file size according to region
        if progress is not None:
            io = BaseProgressCalculator(
                progress, self.path.stat().st_size, "Converting to FASTA"
            )
            io = io.compute_on_read_bytes

        # faidx = self._samtools.fasta_index(reference, regions=regions)
        consensus = self._samtools.consensus(input, output, io=io)
        consensus.wait()
        return output

    def _to_alignment_map(self, target: FileType, regions, progress):
        suffixes = self.path.suffixes.copy()
        if len(suffixes) == 0:
            suffixes = [None]
        reference = None
        if target == FileType.BAM:
            suffixes[-1] = ".bam"
        elif target == FileType.CRAM:
            suffixes[-1] = ".cram"
            if self.file_info.reference_genome.ready_reference is None:
                raise RuntimeError(
                    "Reference genome was not found but is mandatory for CRAM files."
                )
            reference = self.file_info.reference_genome.ready_reference.fasta
        elif target == FileType.SAM:
            suffixes[-1] = ".sam"
        output = self.path.with_name(self.path.stem + "".join(suffixes))

        io = None
        # TODO: get a percentage of input read file size according to region
        if progress is not None:
            io = BaseProgressCalculator(
                progress, self.path.stat().st_size, f"Converting to {target.name}"
            )
            io = io.compute_on_read_bytes

        return self._samtools.view(
            self.path,
            target_format=target,
            output=output,
            regions=regions,
            cram_reference=reference,
            io=io,
        )

    def convert(self, target: FileType, regions=None, progress=None):
        if self.file_info.file_type == target:
            raise ValueError("Target and source file type for conversion are identical")
        if target == FileType.FASTA:
            return self._to_fasta(regions, progress)
        if target == FileType.FASTQ:
            return self._to_fastq(progress)
        return self._to_alignment_map(target, regions, progress)

    def load_meta(self):
        try:
            if self.meta_file.exists():
                with self.meta_file.open("rb") as f:
                    return pickle.load(f)
        except Exception as e:
            logger.error(
                f"Error when loading meta-information for {self.meta_file.name}: {e!s}"
            )
        return None

    def save_meta(self, file_info=None):
        try:
            if file_info is None:
                file_info = self.file_info
            with self.meta_file.open("wb") as f:
                pickle.dump(file_info, f)
        except Exception as e:
            logger.error(
                f"Error when saving meta-information for {self.meta_file.name}: {e!s}"
            )

    def _initialize_file_info(self):
        meta = self.load_meta()
        if meta is not None:
            return meta
        file_info = AlignmentMapFileInfo()
        file_info.path = self.path
        file_info.file_type = AlignmentMapFile.SUPPORTED_FILES[self.path.suffix.lower()]
        file_info.sorted = self.header.metadata.sorted
        file_info.name_type_mtdna = self.header.mtdna_name_type()
        file_info.name_type_chromosomes = self.header.chromosome_name_type()
        file_info.sequence_count = self.header.sequence_count()
        file_info.indexed = self._indexed(file_info.file_type)
        file_info.gender = Gender.Unknown
        file_info.reference_genome = self._repo.find(
            list(self.header.sequences.values())
        )
        file_info.mitochondrial_dna_model = self.get_mitochondrial_dna_type(
            file_info.reference_genome
        )

        # Compute IndexStats automatically only if it's inexpensive to do so.
        # Otherwise, let the caller explicitly request them.
        inexpensive_index_stats = file_info.indexed and file_info.file_type not in [
            FileType.CRAM,
            FileType.SAM,
        ]

        if inexpensive_index_stats:
            indexed_stats = IndexStatsCalculator(self.path)
            file_info.index_stats = indexed_stats.get_stats()
            file_info.gender = self.get_gender(file_info.index_stats)

        # If file is not sorted computing AlignmentStats is expensive.
        # Let the caller request them.
        if file_info.sorted == Sorting.Coordinate:
            is_cram = file_info.file_type == FileType.CRAM
            has_reference = file_info.reference_genome.ready_reference is not None
            if is_cram and has_reference or not is_cram:
                calculator = AlignmentStatsCalculator(file_info)
                file_info.alignment_stats = calculator.get_stats()
        self.save_meta(file_info)
        return file_info

    def get_mitochondrial_dna_type(self, reference: Reference):
        mitochondrial_dna_model = MitochondrialModelType.Unknown
        if reference.status == ReferenceStatus.Available:
            matching = reference.matching[0]
            if matching.mitochondrial_model is not None:
                mitochondrial_dna_model = MitochondrialModelType[
                    matching.mitochondrial_model
                ]
        return mitochondrial_dna_model

    def _indexed(self, type=None):
        if type is None:
            type = self.file_info.file_type

        file = str(self.path)
        if type == FileType.BAM:
            return Path(file + ".bai").exists()
        elif type == FileType.CRAM:
            return Path(file + ".crai").exists()
        return False

    def get_gender(self, stats: list[SequenceStatistics]):
        x_stats = [x for x in stats if x.type == SequenceType.X]
        y_stats = [x for x in stats if x.type == SequenceType.Y]
        x_length = 0
        y_length = 0

        if len(x_stats) != 0 and len(x_stats) != 1:
            return Gender.Unknown

        if len(y_stats) != 0 and len(y_stats) != 1:
            return Gender.Unknown

        if len(x_stats) == 1:
            x_length = x_stats[0].mapped + x_stats[0].unmapped
        if len(y_stats) == 1:
            y_length = y_stats[0].mapped + y_stats[0].unmapped

        if x_length == 0 and y_length == 0:
            return Gender.Unknown
        elif y_length == 0 or (x_length / y_length) > 20:
            return Gender.Female
        else:
            return Gender.Male



import json
import typing

from helix.configuration import MANAGER_CFG
from helix.data.build import Build
from helix.data.genome import Genome
from helix.data.sequence import Sequence
from helix.data.source import Source


class MetadataLoader:
    """Manage metadata related to reference genomes.

    Raises:
        FileNotFoundError: _description_
        FileNotFoundError: _description_
        FileNotFoundError: _description_

    Returns:
        _type_: _description_
    """

    class _CircularReferenceEncoder(json.JSONEncoder):
        """Helps with the serialization of metadata, ensuring
        that is serialized only what's really needed."""

        def __init__(self, **kwargs) -> None:
            self.seen = set()
            super().__init__()

        def default(self, obj):
            if hasattr(obj, "__dict__"):
                obj_dict = obj.__dict__.copy()
                self.seen.add(id(obj))
                for key, value in obj.__dict__.items():
                    if id(value) in self.seen:
                        del obj_dict[key]
                    elif value is None:
                        del obj_dict[key]
                    else:
                        # Trivial types (i.e., str) may share the same id()
                        # if they are identical. This will break the serialization
                        # process. Let's add to the seen set an object only if it is
                        # non trivial.
                        if hasattr(value, "__dict__"):
                            self.seen.add(id(value))
                return {k: v for k, v in obj_dict.items() if not k.startswith("_")}
            return super().default(obj)

    def __init__(self, config=MANAGER_CFG.REPOSITORY):
        self._config = config
        self.genome_root = self._config.genomes
        self.references_path = self._config.metadata.joinpath("references.json")
        self.sources_path = self._config.metadata.joinpath("sources.json")

        if not self.genome_root.is_dir():
            raise FileNotFoundError(
                f"Unable to find {str(self.genome_root)} or is not a directory."
            )
        if not self.references_path.is_file():
            raise FileNotFoundError(
                f"Unable to find {str(self.references_path)} or is not a file."
            )
        if not self.sources_path.is_file():
            raise FileNotFoundError(
                f"Unable to find {str(self.sources_path)} or is not a file."
            )

        self.source_meta: dict[str, Source] = self._load_sources()

    def _load_sources(self):
        with self.sources_path.open("rt") as f:
            decoded = json.load(f)
        sources = {}
        for item in decoded:
            source = Source(**item)
            sources[source.name] = source
            builds = {}
            if source.builds is None:
                continue
            for build in source.builds:
                build = Build(**build)
                builds[build.name] = build
            source.builds = builds
        return sources

    def load(self) -> typing.List[Genome]:
        with self.references_path.open("rt") as f:
            decoded = json.load(f)
        genomes = []
        for genome in [Genome(**x) for x in decoded]:
            genomes.append(genome)
            genome.parent_folder = self.genome_root

            if genome.source in self.source_meta:
                genome.parent = self.source_meta[genome.source]
            else:
                genome.parent = Source(
                    genome.source, [], [], "Metadata for this source is not available."
                )
            if genome.sequences is None:
                continue
            genome.sequences = [Sequence(**x) for x in genome.sequences]
            for sequence in genome.sequences:
                sequence.parent = genome
        return genomes

    def save(self, source: typing.List[Genome]):
        with self.references_path.open("wt") as f:
            json.dump(
                source,
                f,
                cls=MetadataLoader._CircularReferenceEncoder,
                default=lambda o: o.__dict__,
            )



class LetterRun:
    """Represent a single run of a specific letter"""

    def __init__(self) -> None:
        self.start = None
        self.length = None

    def open(self, position: int):
        self.start = position

    def close(self, length: int):
        self.length = length



import collections
import gzip
import logging
import re
import sys
import typing

from helix.alignment_map.alignment_map_header import AlignmentMapHeader
from helix.fasta.letter_run_collection import LetterRunCollection

try:
    import tqdm
except Exception:
    tqdm = None

from helix.reference.genome_metadata_loader import Genome


class FASTALetterCounter:
    def __init__(self, genome: Genome):
        self._bases_progressbar = None
        self.genome = genome
        self._sequences_progressbar = None
        self._bases_progressbar = None
        if not self.genome.dict.exists():
            raise RuntimeError(f"Unable to find dictionary in {self.genome.dict.name}.")
        self._dict: AlignmentMapHeader = AlignmentMapHeader.load_from_file(
            self.genome.dict
        )
        self._support_multiline_progress = "win32" == sys.platform

    @property
    def model_name(self):
        return self.genome.fasta

    def _progress(
        self, name, current_sequence, total_sequences, current_position, total_position
    ):
        if tqdm is not None:
            # Windows Prompt does not support nested progress bar and there's nothing
            # that can be done about that. On Windows, just keep only one progress bar.
            if self._sequences_progressbar is None and self._support_multiline_progress:
                self._sequences_progressbar = tqdm.tqdm(
                    total=total_sequences, desc="Sequences"
                )
            if self._bases_progressbar is None:
                self._bases_progressbar = tqdm.tqdm(
                    total=total_position, desc=f"{name}"
                )

            if current_sequence != 0 and self._support_multiline_progress:
                self._sequences_progressbar.update(current_sequence)
            if current_position != 0:
                self._bases_progressbar.update(current_position)

    def _sequence_from_line(self, line: str) -> LetterRunCollection:
        # Processing the opening of a new sequence in a FASTA file.
        # It's usually a line that begins with '>' followed by the name
        # of the sequence.
        sequence_name = line.split()[0][1:]
        if sequence_name not in self._dict.sequences:
            raise ValueError(f"Sequence {sequence_name} is not present in dictionary.")
        sequence_length = self._dict.sequences[sequence_name].length
        return LetterRunCollection(sequence_name, sequence_length)

    def _process_file(
        self,
        fp: typing.TextIO,
        letter: str,
        progress: typing.Callable[[int, int, int, int], None],
    ) -> typing.List[LetterRunCollection]:
        sequences = collections.OrderedDict()
        pattern = re.compile(rf"{letter}+")

        position = None
        current_sequence = None

        for line in fp:
            line = line.rstrip("\n")

            # Comment: skip
            if len(line) == 0 or line[0] == "#":
                continue

            # Check if processing a .fastq file by mistake.
            if line[0] == "+":
                raise RuntimeError(
                    "Expected a FASTA reference model, got a FASTQ sequencer data."
                )

            if line[0] == ">":
                if self._bases_progressbar is not None:
                    self._bases_progressbar.close()
                    self._bases_progressbar = None
                # New sequence found. Close old sequence if open.
                if current_sequence is not None:
                    current_sequence.end(position)

                current_sequence = self._sequence_from_line(line)
                if current_sequence.name in sequences:
                    raise RuntimeError(
                        f"Found a duplicated sequence: {current_sequence.name}"
                    )
                logging.debug(
                    f"{self.genome.fasta.name}: Processing sequence {current_sequence.name}"
                )
                sequences[current_sequence.name] = current_sequence
                if progress is not None:
                    progress(
                        current_sequence.name,
                        1,
                        len(self._dict.sequences),
                        0,
                        current_sequence.length,
                    )
                position = 0
                continue

            matches = list(pattern.finditer(line))

            # No Ns found in this line: close the open run (if any).
            if len(matches) == 0:
                if current_sequence.is_run_open():
                    current_sequence.close_run(position)

            for match in matches:
                if match.start() == 0:
                    # The Ns starts at the beginning of the line.
                    # If a run is open, keep open: is continuing from the previous line.
                    # Otherwise open a new one.
                    if not current_sequence.is_run_open():
                        current_sequence.open_run(position)
                else:
                    # There are Ns but they don't start at the beginning of the line.
                    # Close the run (if open) and re-open it: this is a new run of Ns.
                    if current_sequence.is_run_open():
                        current_sequence.close_run(position)
                    current_sequence.open_run(match.start() + position)

                # The run is ending before the end of the line. Close the open run (if any).
                if match.end() < len(line):
                    current_sequence.close_run(match.end() + position)
            position += len(line)
            if progress is not None:
                progress(
                    current_sequence.name,
                    0,
                    len(self._dict.sequences),
                    len(line),
                    current_sequence.length,
                )

        # File is terminated: close the current sequence and the open run (if any).
        if current_sequence is None:
            raise RuntimeError("Found the end of the file but no sequences found.")
        current_sequence.end(position)
        return list(sequences.values())

    def count_letters(
        self,
        letter: str = "N",
        progress: typing.Callable[[str, int, int, int, int], None] = False,
    ) -> typing.List[LetterRunCollection]:
        if progress is False:
            progress = self._progress

        with gzip.open(self.genome.fasta, "rt") as f:
            sequences = self._process_file(f, letter, progress)
        return sequences



from helix.data.microarray_converter import (
    MicroarrayConverterExtensions,
    MicroarrayConverterTarget,
)

TARGET_FORMATTER_MAP: dict[
    MicroarrayConverterTarget, tuple[callable, MicroarrayConverterExtensions]
] = dict()


def target(
    target: MicroarrayConverterTarget, extension=MicroarrayConverterExtensions.TXT
):
    """Simple decorator that it binds a microarray target with a function.

    Args:
        target (MicroarrayConverterTarget): Target to bind to the function
        it's decorating.
    """

    def inner_decorator(func):
        TARGET_FORMATTER_MAP[target] = (func, extension)
        return func

    return inner_decorator


class MicroarrayLineFormatter:
    @target(MicroarrayConverterTarget.FTDNA_v1, MicroarrayConverterExtensions.CSV)
    @target(MicroarrayConverterTarget.MyHeritage_v2, MicroarrayConverterExtensions.CSV)
    def to_quote_comma_separated(id, chromosome, position, result):
        return f'"{id}","{chromosome}","{position}","{result}"\n'

    @target(MicroarrayConverterTarget.FTDNA_v3, MicroarrayConverterExtensions.CSV)
    def to_comma_separated(id, chromosome, position, result):
        return f"{id},{chromosome},{position},{result}\n"

    @target(MicroarrayConverterTarget.LivingDNA_v1)
    @target(MicroarrayConverterTarget.LivingDNA_v2)
    @target(MicroarrayConverterTarget.TwentyThreeAndMe_SNPs_API)
    @target(MicroarrayConverterTarget.TwentyThreeAndMe_v3)
    @target(MicroarrayConverterTarget.TwentyThreeAndMe_v35)
    @target(MicroarrayConverterTarget.TwentyThreeAndMe_v4)
    @target(MicroarrayConverterTarget.TwentyThreeAndMe_v5)
    def to_tab_separated(id, chromosome, position, result):
        return f"{id}\t{chromosome}\t{position}\t{result}\n"

    @target(MicroarrayConverterTarget.Ancestry_v1)
    @target(MicroarrayConverterTarget.Ancestry_v2)
    def handle_ancestry(id, chromosome, position, result):
        map_results = {"--": "00", "CT": "TC", "GT": "TG"}
        if result in map_results:
            result = map_results[result]
        return f"{id}\t{chromosome}\t{position}\t{result[0]}\t{result[1]}\n"

    @target(MicroarrayConverterTarget.MyHeritage_v1, MicroarrayConverterExtensions.CSV)
    def handle_my_heritage_v1(id, chromosome, position, result):
        map_results = {"CT": "TC", "GT": "TG"}
        if result in map_results:
            result = map_results[result]
        return MicroarrayLineFormatter.to_quote_comma_separated(
            id, chromosome, position, result
        )

    @target(MicroarrayConverterTarget.FTDNA_v2, MicroarrayConverterExtensions.CSV)
    def handle_ftdna_v2(id, chromosome, position, result):
        output = ""
        if id == "rs5939319":
            output += "RSID,CHROMOSOME,POSITION,RESULT\n"
        return output + MicroarrayLineFormatter.to_quote_comma_separated(
            id, chromosome, position, result
        )



import json

from helix.configuration import MANAGER_CFG
from helix.data.mitochondrial_model import MitochondrialModel
from helix.data.mitochondrial_model_type import MitochondrialModelType
from helix.data.sequence import Sequence


class MtDNA:
    def __init__(
        self,
        config=MANAGER_CFG.REPOSITORY,
    ) -> None:
        self._mtdna_json = config.metadata.joinpath("mtdna.json")
        self.data: list[MitochondrialModel] = self._load()

    def _load(self):
        json_models = None
        with self._mtdna_json.open("rt") as p:
            json_models = json.load(p)

        models = []
        for model in json_models:
            sequence = Sequence(model["name"], int(model["length"]), model["md5"])
            model = MitochondrialModel(
                sequence, model["url"], MitochondrialModelType[model["name"]]
            )
            models.append(model)

        return models

    def get_by_md5(self, md5):
        for model in self.data:
            if model.sequence.md5 == md5:
                return model
        return None

    def get_by_length(self, length: int):
        for model in self.data:
            if model.sequence.length == length:
                return model
        return None

    def get_by_type(self, type):
        for model in self.data:
            if model.type == type:
                return model
        return None



from helix.alignment_map.index_stats_calculator import SequenceStatistics
from helix.data.sequence_type import SequenceType
from helix.data.tabular_data import TabularData, TabularDataRow
from helix.naming.converter import Converter


class IndexStatsAdapter:
    def adapt(stats: list[SequenceStatistics]):
        header = [
            "Canonical Name",
            "Sequence Type",
            "Reference length",
            "Mapped length",
            "Unmapped length",
        ]
        grouped_stats = []
        grouped_others = TabularDataRow(
            "Others",
            [
                "Others",
                "Others",
                sum(x.reference_length for x in stats if x.type == SequenceType.Other),
                sum(x.mapped for x in stats if x.type == SequenceType.Other),
                sum(x.unmapped for x in stats if x.type == SequenceType.Other),
            ],
        )

        percentage_mapped = ""
        percentage_unmapped = ""
        if grouped_others.columns[2] != 0:
            percentage_mapped = (
                f" ({(grouped_others.columns[3]/grouped_others.columns[2])*100:.1f}%)"
            )
            percentage_unmapped = (
                f" ({(grouped_others.columns[4]/grouped_others.columns[2])*100:.1f}%)"
            )

        grouped_others.columns[3] = f"{grouped_others.columns[3]}{percentage_mapped}"
        grouped_others.columns[4] = f"{grouped_others.columns[4]}{percentage_unmapped}"
        grouped_others.columns[2] = str(grouped_others.columns[2])

        for stat in stats:
            if stat.type == SequenceType.Other:
                continue

            percentage_unmapped = ""
            percentage_mapped = ""
            if stat.reference_length != 0:
                percentage_unmapped = (
                    f" ({(stat.unmapped/stat.reference_length)*100:.1f}%)"
                )
                percentage_mapped = f" ({(stat.mapped/stat.reference_length)*100:.1f}%)"

            grouped_stats.append(
                TabularDataRow(
                    stat.name,
                    [
                        Converter.canonicalize(stat.name),
                        stat.type.name,
                        str(stat.reference_length),
                        f"{stat.mapped}{percentage_mapped}",
                        f"{stat.unmapped}{percentage_unmapped}",
                    ],
                )
            )
        grouped_stats.append(grouped_others)

        return TabularData(header, grouped_stats)



from helix.naming.converter import Converter


class Sequence:
    def __init__(self, name: str, length: int, md5: str = None, parent=None) -> None:
        self.name = name
        self.length = length
        self.md5 = md5
        self.__parent = parent
        self.__type = None
        self.__canonic_name = None

    @property
    def type(self):
        if self.__type is None:
            self.__type = Converter.get_type(self.name)
        return self.__type

    @property
    def canonic_name(self):
        if self.__canonic_name is None:
            self.__canonic_name = Converter.canonicalize(self.name)
        return self.__canonic_name

    @property
    def parent(self):
        return self.__parent

    @parent.setter
    def parent(self, value):
        self.__parent = value

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        if self.md5 is None:
            return f"{self.name}, {self.length}bp"
        else:
            max_length = min(len(self.md5), 6)
            return f"{self.name}, {self.length}bp, ..{self.md5[-max_length:]}"



from PySide6.QtCore import QMetaObject, Qt
from PySide6.QtWidgets import (
    QAbstractItemView,
    QComboBox,
    QDialog,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    # QHeaderView,
)

from helix.data.tabular_data import TabularData


class TableDialog(QDialog):
    def __init__(self, title="Dialog", parent=None, f=Qt.WindowType.Dialog) -> None:
        super().__init__(parent, f)
        self.setObjectName(title)
        self.setWindowTitle(title)
        self.resize(640, 480)

        self.tableWidget = QTableWidget(self)
        # self.tableWidget.horizontalHeader().setSectionResizeMode(
        #     QHeaderView.ResizeMode.ResizeToContents
        # )
        self.tableWidget.setEditTriggers(QAbstractItemView.NoEditTriggers)

        self.verticalLayout = QVBoxLayout(self)
        self.verticalLayout.addWidget(self.tableWidget)

        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        QMetaObject.connectSlotsByName(self)

    def set_data(self, data: TabularData):
        self.tableWidget.setRowCount(len(data.rows))
        self.tableWidget.setColumnCount(max(len(x.columns) for x in data.rows))
        self.tableWidget.horizontalHeader().setStretchLastSection(True)
        if data.horizontal_header:
            for index, header in enumerate(data.horizontal_header):
                self.tableWidget.setHorizontalHeaderItem(
                    index, QTableWidgetItem(header)
                )
        else:
            self.tableWidget.horizontalHeader().hide()

        vertical_header = False
        for row_index, row in enumerate(data.rows):
            if row.vertical_header is not None:
                self.tableWidget.setVerticalHeaderItem(
                    row_index, QTableWidgetItem(row.vertical_header)
                )
                vertical_header = True
            for col_index, text in enumerate(row.columns):
                self.tableWidget.setItem(row_index, col_index, QTableWidgetItem(text))
        if not vertical_header:
            self.tableWidget.verticalHeader().hide()


class ListTableDialog(TableDialog):
    """A simple dialog that is showing a list and a table"""

    def __init__(self, title="Dialog", parent=None, f=Qt.WindowType.Dialog) -> None:
        super().__init__(title, parent, f)
        self.itemList = QComboBox(self)
        self.verticalLayout.insertWidget(0, self.itemList)
        self.data = None
        self.itemList.currentIndexChanged.connect(self.index_changed)

    def set_data(self, data: dict[str, TabularData]):
        self.itemList.addItems(list(data.keys()))
        self.data = data
        self.index_changed(0)

    def index_changed(self, index):
        if self.data is None:
            return
        text = self.itemList.itemText(index)
        super().set_data(self.data[text])



import enum
from pathlib import Path

from helix.utility.external import External


class FileType(enum.Enum):
    ZIP = 0
    BZIP = 1
    SEVENZIP = 2
    # GZIP is not used ATM as it's not reliable.
    GZIP = 3
    BGZIP = 4
    RAZF_GZIP = 5
    DECOMPRESSED = 6


class FileTypeChecker:

    _EXT_TO_TYPE = {
        ".7z": FileType.SEVENZIP,
        ".zip": FileType.ZIP,
        ".bz2": FileType.BZIP,
        ".bz": FileType.BZIP,
        ".gz": FileType.RAZF_GZIP,
        ".fa": FileType.DECOMPRESSED,
        ".fasta": FileType.DECOMPRESSED,
        ".fna": FileType.DECOMPRESSED,
    }

    _TYPE_TO_EXT = {
        FileType.SEVENZIP: ".7z",
        FileType.ZIP: ".zip",
        FileType.BZIP: ".bz2",
        FileType.RAZF_GZIP: ".fa.gz",
        FileType.GZIP: ".fa.gz",
        FileType.DECOMPRESSED: ".fa",
    }

    _HTSFILE_TO_TYPE = {
        "BGZF": FileType.BGZIP,
        "gzip": FileType.RAZF_GZIP,
        "RAZF": FileType.RAZF_GZIP,
        "FASTA": FileType.DECOMPRESSED,
    }

    def __init__(self, external: External = External()) -> None:
        self._external = external

    def get_type(self, file: Path) -> FileType:
        """Get a Type starting from a file path.

        Args:
            file (Path): Path of the file to analyze.

        Returns:
            Type | None: Type or None if the type is unknown.
        """
        # Extensions can be wrong or misleading; Use htsfile and
        # eventually fallback on extension.

        file_type = self._external.get_file_type(file)
        for key, value in FileTypeChecker._HTSFILE_TO_TYPE.items():
            if key in file_type:
                return value

        # TODO: starting from .gz in _EXT_TO_TYPE, htsfile should really
        # be able to give something useful. If it isn't the case, it could
        # mean the file is corrupted. Unsure what's the best logic here.
        extension = file.suffix
        if extension in FileTypeChecker._EXT_TO_TYPE:
            return FileTypeChecker._EXT_TO_TYPE[extension]
        return None

    def get_extension(self, file: Path) -> str:
        type = self.get_type(file)
        matches = [x for x in FileTypeChecker._EXT_TO_TYPE.items() if x[1] == type]
        if len(matches) > 0:
            return matches[0][0]
        return None



from helix.data.sequence import Sequence


class AlignmentMapHeaderSequence(Sequence):
    def __init__(
        self, name: str = None, length: int = None, md5: str = None, parent=None
    ) -> None:
        super().__init__(name, length, md5, parent)
        self.uri: str = None
        self.alternate_locus: str = None
        self.molecule_topology: str = None
        self.species: str = None
        self.alternative_names: str = None
        self.genome_assembly_identifier: str = None
        self.description: str = None



from pathlib import Path
from test.utility import MockFile
from unittest.mock import Mock, patch

from helix.microarray.raw_file import RawEntry, RawFile


def test_raw_entry():
    sut = RawEntry("rs1", "1", 1, "AA")
    assert sut.template_entry == "rs1\t1\t1\n"


@patch("gzip.open")
@patch("pathlib.Path.open")
def test_compressed_raw_file(mock_open, mock_gzopen):
    lines = []
    file = Mock()
    file.write = lambda x: lines.append(x)

    mock_gzopen.side_effect = [
        MockFile(["#FOO", " rs1 \t 1 \t 1 \t AA \n", "rs2\tX\t123\tBB"])
    ]
    mock_open.side_effect = [file]

    elements: list[RawEntry] = []

    for item in RawFile(Path("test.txt.gz")):
        elements.append(item)

    assert len(elements) == 2
    assert elements[0].id == "rs1"
    assert elements[0].chromosome == "1"
    assert elements[0].position == 1
    assert elements[0].result == "AA"
    assert elements[1].id == "rs2"
    assert elements[1].chromosome == "X"
    assert elements[1].position == 123
    assert elements[1].result == "BB"


# @patch('builtins.open')
# @patch('pathlib.Path.open')
# def test_compressed_raw_file(mock_open, mock_gzopen):
#     lines = []
#     file = Mock()
#     file.write = lambda x: lines.append(x)

#     mock_gzopen.side_effect = [MockFile(["#FOO", " rs1 \t 1 \t 1 \t AA \n", "rs2\tX\t123\tBB"])]
#     mock_open.side_effect = [file]

#     elements : list[RawEntry] = []

#     for item in RawFile(Path("test.txt")):
#         elements.append(item)

#     assert len(elements) == 2
#     assert elements[0].id == "rs1"
#     assert elements[0].chromosome == "1"
#     assert elements[0].position == 1
#     assert elements[0].result == "AA"
#     assert elements[1].id == "rs2"
#     assert elements[1].chromosome == "X"
#     assert elements[1].position == 123
#     assert elements[1].result == "BB"



import logging
import os
import signal
from subprocess import Popen
from threading import Thread

import debugpy


class SimpleWorker(Thread):
    def __init__(self, function, *args, **kwargs) -> None:
        super().__init__()
        self.function = function
        self.process = None
        self._kwargs = kwargs
        self._args = args
        self.start()

    def run(self):
        if debugpy.is_client_connected():
            debugpy.debug_this_thread()
        self.process = self.function(*self._args, **self._kwargs)
        if self.process is not None and isinstance(self.process, Popen):
            self.process.communicate()

    def kill(self):
        process = self.process
        if process is None:
            logging.error("Trying to kill a SimpleWorker but there's nothing to kill.")
            return
        try:
            if hasattr(os.sys, "winver") or not isinstance(process, Popen):
                process.kill()
            else:
                process.send_signal(signal.SIGTERM)
        except Exception as e:
            logging.error(f"Error when killing process {e!s}")



import enum
import inspect
import logging
import subprocess
from logging.handlers import TimedRotatingFileHandler
from subprocess import Popen

import helix.utility.external
from helix.configuration import MANAGER_CFG
from helix.utility.external import External

logger = logging.getLogger(__name__)


class PrerequisiteIssueSeverity(enum.Enum):
    INFO = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()


class PrerequisiteIssueType(enum.Enum):
    CONFIG_ISSUE = enum.auto()
    MISSING_BINARY = enum.auto()
    BINARY_ERROR = enum.auto()


class PrerequisiteIssue:
    def __init__(
        self,
        severity: PrerequisiteIssueSeverity,
        type: PrerequisiteIssueType,
        description: str,
    ) -> None:
        self.severity = severity
        self.type = type
        self.description = description

    def __str__(self) -> str:
        return f"{self.severity.name}: {self.type.name}, {self.description}"

    def __repr__(self) -> str:
        return self.__str__()


class CheckPrerequisites:
    def __init__(self) -> None:
        self._repo = MANAGER_CFG.REPOSITORY
        self._ext = MANAGER_CFG.EXTERNAL
        self._gen = MANAGER_CFG.GENERAL
        self._config_logging()

    def _config_logging(self):
        if not self._repo.log_path.exists():
            self._repo.log_path.mkdir(parents=True, exist_ok=True)
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            logger.removeHandler(handler)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

        log_file = self._repo.log_path.joinpath("app.log")
        handler = TimedRotatingFileHandler(
            log_file, when="midnight", interval=1, backupCount=10
        )
        handler.suffix = "%Y-%m-%d"
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        return logger

    def check_prerequisites(self) -> list[PrerequisiteIssue]:
        # TODO: this sucks
        # TODO: empty config var should fallback on defaults with an INFO msg
        logger.info("Self test started.")
        old_level = logging.getLogger(helix.utility.external.__name__).level
        logging.getLogger(helix.utility.external.__name__).setLevel(logging.INFO)
        missing = {
            x: []
            for x in [
                PrerequisiteIssueSeverity.ERROR,
                PrerequisiteIssueSeverity.WARNING,
                PrerequisiteIssueSeverity.INFO,
            ]
        }
        if self._repo.metadata is None:
            missing[PrerequisiteIssueSeverity.ERROR].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    "Metadata field is empty in repository config",
                )
            )
        elif not self._repo.metadata.exists():
            missing[PrerequisiteIssueSeverity.ERROR].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    f"Metadata folder does not exist: {self._repo.metadata!s}",
                )
            )

        if self._repo.genomes is None:
            missing[PrerequisiteIssueSeverity.ERROR].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    "Repository field is empty in repository config",
                )
            )
        elif not self._repo.genomes.exists():
            missing[PrerequisiteIssueSeverity.ERROR].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    f"Repository folder does not exist: {self._repo.genomes!s}",
                )
            )

        if self._repo.temporary is None:
            missing[PrerequisiteIssueSeverity.ERROR].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    "Temporary folder field is empty in repository config",
                )
            )
        elif not self._repo.temporary.exists():
            self._repo.temporary.mkdir(parents=True)
            missing[PrerequisiteIssueSeverity.INFO].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.INFO,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    f"Temporary folder does not exist present. Creating: {self._repo.temporary!s}",
                )
            )

        if self._ext.root is None:
            missing[PrerequisiteIssueSeverity.ERROR].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    f"Root folder does not exist: {self._ext.root!s}",
                )
            )
        elif not self._repo.temporary.exists():
            missing[PrerequisiteIssueSeverity.INFO].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    f"Temp folder does not exist: {self._ext.root!s}. Creating",
                )
            )
            self._repo.temporary.mkdir(parents=True, exist_ok=True)

        if self._ext.threads is None:
            missing[PrerequisiteIssueSeverity.ERROR].append(
                PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.CONFIG_ISSUE,
                    "Threads field is empty in external config",
                )
            )

        ext = External()
        members = inspect.getmembers(External, lambda x: inspect.isfunction(x))
        decorated = [x[1] for x in members if hasattr(x[1], "wrapper")]
        for f in decorated:
            try:
                result: Popen = f(
                    ext,
                    stdout=subprocess.PIPE,
                    stdin=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                result.communicate()
            except FileNotFoundError:
                m = PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.MISSING_BINARY,
                    f.__name__,
                )
                missing[PrerequisiteIssueSeverity.ERROR].append(m)
            except Exception as e:
                m = PrerequisiteIssue(
                    PrerequisiteIssueSeverity.ERROR,
                    PrerequisiteIssueType.BINARY_ERROR,
                    f"{f.__name__} : {e!s}",
                )
                missing[PrerequisiteIssueSeverity.ERROR].append(m)
        logging.getLogger(helix.utility.external.__name__).setLevel(old_level)
        logger.info(
            f"Self test ended: {len(missing[PrerequisiteIssueSeverity.ERROR])} errors, "
            + f"{len(missing[PrerequisiteIssueSeverity.WARNING])} warning, "
            + f"{len(missing[PrerequisiteIssueSeverity.INFO])} info."
        )
        return missing


if __name__ == "__main__":
    failing = CheckPrerequisites().check_prerequisites()
    if len(failing) == 0:
        print("All requirements are satisfied")
    else:
        for req in failing:
            print(req)



import collections
import math
import typing

from helix.fasta.letter_run_collection import LetterRunCollection


class LetterRunBuckets:
    """Represent a collection of run of letters partitioned into subsequences of fixed length called buckets"""

    def __init__(
        self,
        sequence: LetterRunCollection,
        buckets_number: int,
        long_run_threshold: int,
        items: typing.OrderedDict[int, int] = None,
    ) -> None:
        self._buckets_number = buckets_number
        self._long_run_threshold: int = long_run_threshold
        self._sequence: LetterRunCollection = sequence
        self.buckets = items

        if self.buckets is None:
            self.buckets = self._make_buckets()

    def _make_buckets(self) -> typing.OrderedDict[int, int]:
        if self._buckets_number > self._sequence.length:
            # Can't have less than 1 N per bucket.
            # This can happen in short sequence (i.e., decoy).
            return collections.OrderedDict()

        buckets = collections.OrderedDict()
        bucket_size = int(math.floor(self._sequence.length / self._buckets_number))

        for run in self._sequence.filter(
            lambda x: x.length >= self._long_run_threshold
        ):
            # Determine how many buckets this run is spanning
            start = run.start
            end = run.start + run.length
            index_bucket_start = int(math.floor(start / bucket_size))
            index_bucket_end = int(math.floor(end / bucket_size))

            # Iterate over each bucket determining how many elements are in
            # each bucket.
            for bucket in range(index_bucket_start, index_bucket_end + 1):
                start_bucket_offset = max(bucket * bucket_size, start)
                end_bucket_offset = min(bucket * bucket_size + bucket_size - 1, end)
                runs_count = end_bucket_offset - start_bucket_offset
                if runs_count < 1:
                    continue
                if bucket not in buckets:
                    buckets[bucket] = 0
                buckets[bucket] += runs_count
        return buckets



import typing


class Build:
    def __init__(
        self,
        name: str,
        description=None,
        urls: typing.List[str] = None,
    ) -> None:
        self.name = name
        self.description = description
        self.urls = urls



from helix.data.coverage_stats import CoverageStats, DepthBin
from helix.data.tabular_data import TabularData, TabularDataRow


class CoverageStatsAdapter:
    def adapt(input_data: list[CoverageStats]):
        rows = []
        for sequence_stats in input_data:
            row = [
                sequence_stats.sequence_name,
                str(sequence_stats.bin_entries[DepthBin.Zero]),
                str(sequence_stats.bin_entries[DepthBin.Between0And3]),
                str(sequence_stats.bin_entries[DepthBin.Between3And7]),
                str(sequence_stats.bin_entries[DepthBin.MoreThan7]),
                str(sequence_stats.all_average),
                str(sequence_stats.non_zero_average),
            ]
            rows.append(TabularDataRow(None, row))
        horizontal_header = [
            "Sequence Name",
            "# Reads: 0",
            "# Reads: 0-3",
            "# Reads: 3-7",
            "# Reads: >7",
            "Average",
            "Non-zero average",
        ]
        return TabularData(horizontal_header, rows)



import hashlib
import typing
from pathlib import Path

from helix.data.build import Build
from helix.data.sequence import Sequence


class Genome:
    def __init__(
        self,
        fasta_url: str,
        fai_url: str = None,
        gzi_url: str = None,
        suffix: str = None,
        build: str = None,
        source: str = None,
        mitochondrial_name_type: str = None,
        nuclear_name_type: str = None,
        sequences: typing.List[Sequence] = None,
        description: str = None,
        download_size: int = None,
        decompressed_size: int = None,
        bgzip_size: int = None,
        downloaded_md5: str = None,
        decompressed_md5: str = None,
        bgzip_md5: str = None,
        mitochondrial_model=None,
        parent_folder=Path("."),
    ) -> None:
        self.fasta_url = fasta_url
        self.fai_url = fai_url
        self.gzi_url = gzi_url
        self.suffix = suffix
        self.build = build
        self.source = source
        self.mitochondrial_name_type = mitochondrial_name_type
        self.nuclear_name_type = nuclear_name_type
        self.sequences = sequences
        self.description = description
        self.download_size = download_size
        self.decompressed_size = decompressed_size
        self.bgzip_size = bgzip_size
        self.downloaded_md5 = downloaded_md5
        self.decompressed_md5 = decompressed_md5
        self.bgzip_md5 = bgzip_md5
        self.mitochondrial_model = mitochondrial_model
        # Not serialized as it depends on the config
        self.__parent_folder = parent_folder
        # Not serialized as it's populated at runtime
        self.__parent: Build = None

    @property
    def parent_folder(self):
        return self.__parent_folder

    @parent_folder.setter
    def parent_folder(self, parent: Path):
        if isinstance(parent, str):
            parent = Path(parent)
        self.__parent_folder = parent

    @property
    def parent(self):
        return self.__parent

    @parent.setter
    def parent(self, value):
        self.__parent = value

    @property
    def url_hash(self) -> str:
        url = bytes(self.fasta_url, encoding="utf8")
        return hashlib.md5(url).hexdigest()

    @property
    def all(self) -> typing.List[Path]:
        return [
            self.fasta,
            self.gzi,
            self.nbin,
            self.nbuc,
            self.bed,
            self.fai,
            self.dict,
        ]

    @property
    def no_extensions(self) -> Path:
        if self.suffix is not None:
            filename = f"{self.build}_{self.suffix}_{self.source}_{self.url_hash[0:8]}"
        else:
            filename = f"{self.build}_{self.source}_{self.url_hash[0:8]}"
        return self.__parent_folder.joinpath(filename)

    @property
    def fasta(self) -> Path:
        return Path(str(self.no_extensions) + ".fa.gz")

    @property
    def decompressed(self) -> Path:
        return Path(str(self.no_extensions) + ".fa")

    @property
    def gzi(self) -> Path:
        return Path(str(self.fasta) + ".gzi")

    @property
    def fai(self) -> Path:
        return Path(str(self.fasta) + ".fai")

    @property
    def nbin(self) -> Path:
        return Path(str(self.no_extensions) + "_nbin.csv")

    @property
    def nbuc(self) -> Path:
        return Path(str(self.no_extensions) + "_nbuc.csv")

    @property
    def bed(self) -> Path:
        return Path(str(self.no_extensions) + "_nreg.bed")

    @property
    def dict(self) -> Path:
        return Path(str(self.no_extensions) + ".dict")

    def __repr__(self) -> str:
        return str(self.no_extensions.name)

    def __str__(self) -> str:
        return self.__repr__()



import argparse
import logging
import time
from pathlib import Path

from helix.data.genome import Genome
from helix.fasta.fasta_letter_counter import FASTALetterCounter
from helix.fasta.fasta_stats_files import FASTAStatsFiles
from helix.utility.samtools import Samtools

if __name__ == "__main__":
    parser = argparse.ArgumentParser(__name__)
    parser.add_argument(
        "--reference",
        help="Indicate the path of the bgzip compressed reference genome",
        type=Path,
        default=Path("reference\\genomes\\_hs37d5.fa\\seq1.fa.gz"),
    )
    parser.add_argument(
        "--reference-dict",
        help="Indicate the path of dictionary for the reference genome",
        type=Path,
        default=None,
    )
    parser.add_argument(
        "--long-run-threshold",
        help="Indicate the threshold for long runs to be considered when making .buc files.",
        type=int,
        default=300,
    )
    parser.add_argument(
        "--buckets-number",
        help="Indicate the total number of buckets to be considered when making .buc files.",
        type=int,
        default=1000,
    )
    parser.add_argument(
        "--external",
        help="Indicate the root directory for 3rd parties executables (only needed to generate the .dict file if not already available).",
        type=Path,
        default=None,
    )
    parser.add_argument(
        "--verbose",
        help="Set logging level",
        choices=["INFO", "DEBUG", "WARNING", "ERROR"],
        type=str,
        default="INFO",
    )
    args = parser.parse_args()

    logging.getLogger().setLevel(logging.__dict__[args.verbose])

    start = time.time()
    genome = Genome(args.reference)

    if not genome.dict.exists():
        samtools = Samtools()
        samtools.make_dictionary(args.reference, genome.dict)

    fasta_file = FASTALetterCounter(genome)
    unknown_bases_stats = FASTAStatsFiles(
        fasta_file,
        args.long_run_threshold,
        args.buckets_number,
    )
    unknown_bases_stats.generate_stats()
    end = time.time()
    logging.info(f"Time {end-start}s")



class TabularDataRow:
    def __init__(self, header: str, columns: list[str]) -> None:
        self.vertical_header = header
        self.columns = columns


class TabularData:
    def __init__(
        self, horizontal_header: list[str], rows: list[TabularDataRow]
    ) -> None:
        self.horizontal_header = horizontal_header
        self.rows = rows



from helix.data.tabular_data import TabularData, TabularDataRow
from helix.fasta.reference import Reference


class ReferenceAdapter:
    def adapt(stats: Reference):
        ref_map = {}
        if len(stats.matching) > 0:
            # If We've at least a perfect match,
            # build a table with it
            for match in stats.matching:
                for key, value in stats.reference_map.items():
                    for sequence in value:
                        if sequence.parent == match:
                            if key not in ref_map:
                                ref_map[key] = []
                            ref_map[key].append(sequence)
        else:
            # Otherwise, consider every reference that
            # matches at least by one sequence
            ref_map = stats.reference_map
        parents = set(x.parent for y in ref_map.values() for x in y)
        max_columns = len(parents)
        genome_index_map = {}
        headers = ["Loaded file"]
        max_index = 0
        rows = []
        for key, value in ref_map.items():
            row = [None for x in range(max_columns + 1)]
            row[0] = key
            for sequence in value:
                if sequence.parent not in genome_index_map:
                    genome_index_map[sequence.parent] = max_index
                    headers.append(str(sequence.parent))
                    max_index += 1
                row[genome_index_map[sequence.parent] + 1] = sequence
            rows.append([str(x) for x in row])
        return TabularData(headers, [TabularDataRow(None, x) for x in rows])



import enum


class ChromosomeNameType(enum.Enum):
    Chr = enum.auto()
    Number = enum.auto()
    GenBank = enum.auto()
    RefSeq = enum.auto()
    GenBankT2T = enum.auto()
    RefSeqT2T = enum.auto()
    Unknown = enum.auto()



import pytest

from helix.fasta.letter_run_buckets import LetterRunBuckets
from helix.fasta.letter_run_collection import LetterRunCollection


def test_single_bucket():
    # Arrange
    sequence = LetterRunCollection("Foo", 1000)
    sequence.open_run(0)
    sequence.close_run(100)
    # Act
    # 1000/10 = 100 -> 10 buckets with length of 100
    sut = LetterRunBuckets(sequence, 10, 0).buckets
    # Assert
    assert 0 in sut
    assert len(sut) == 1
    assert sut[0] == 99


def test_more_run_in_single_bucket():
    # Arrange
    sequence = LetterRunCollection("Foo", 1000)
    sequence.open_run(0)
    sequence.close_run(50)

    sequence.open_run(51)
    sequence.close_run(60)
    # Act
    # 1000/10 = 100 -> 10 buckets with length of 100
    sut = LetterRunBuckets(sequence, 10, 0).buckets
    # Assert
    assert 0 in sut
    assert len(sut) == 1
    assert sut[0] == 59


def test_run_in_multiple_buckets():
    # Arrange
    sequence = LetterRunCollection("Foo", 1000)
    sequence.open_run(0)
    sequence.close_run(150)
    # Act
    # 1000/10 = 100 -> 10 buckets with length of 100
    sut = LetterRunBuckets(sequence, 10, 0).buckets
    # Assert
    assert 0 in sut
    assert 1 in sut
    assert len(sut) == 2
    assert sut[0] == 99  # 0 -> 99
    assert sut[1] == 50  # 99 -> 151


def test_too_many_buckets():
    sequence = LetterRunCollection("Foo", 1000)

    sequence.open_run(0)
    sequence.close_run(99)

    sequence.open_run(100)
    sequence.close_run(201)

    sut = LetterRunBuckets(sequence, 1001, 0).buckets
    assert len(sut) == 0


def test_run_end_to_end_divisible():
    sequence = LetterRunCollection("Foo", 1000)

    sequence.open_run(0)
    sequence.close_run(1000)
    sut = LetterRunBuckets(sequence, 10, 0).buckets
    for index in range(10):
        assert index in sut
        assert sut[index] == 99


def test_run_end_to_end_non_divisible():
    sequence = LetterRunCollection("Foo", 1000)

    sequence.open_run(0)
    sequence.close_run(1000)
    sut = LetterRunBuckets(sequence, 3, 0).buckets
    assert 0 in sut
    assert 1 in sut
    assert 2 in sut
    assert 3 in sut

    assert len(sut) == 4
    assert sut[0] == 332
    assert sut[1] == 332
    assert sut[2] == 332
    assert sut[3] == 1



import sys

from PySide6.QtWidgets import QApplication

from helix.gui.main import HelixWindow


def main():
    app = QApplication(sys.argv)
    # Use a less broken UI style
    app.setStyle("fusion")
    widget = HelixWindow()
    widget.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



import typing

from helix.data.build import Build


class Source:
    def __init__(
        self,
        name: str,
        urls: typing.List[str] = None,
        builds: typing.List[Build] = None,
        description: str = None,
    ) -> None:
        self.name = name
        self.builds = builds
        self.description = description
        self.urls = urls

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"{self.name}"



from helix.data.alignment_map.alignment_map_file_info import AlignmentMapFileInfo
from helix.data.tabular_data import TabularData, TabularDataRow


class AlignmentMapFileInfoAdapter:
    def adapt(stats: AlignmentMapFileInfo):
        label_map = {
            "path": "Path",
            "sorted": "Sorted",
            "indexed": "Indexed",
            "file_type": "File type",
            "reference_genome": "Reference genome",
            "content": "Content",
            "mitochondrial_dna_type": "Mitochondrial DNA type",
            "name_type_chromosomes": "Name type chromosomes",
            "name_type_mtdna": "Name type mtDNA",
            "sequence_count": "Sequence count",
            "gender": "Gender",
        }
        size = stats.path.stat().st_size
        size /= 1024**3
        size = f"{size:.1f} GB"

        data = {x: None for x in label_map.values()}
        data["Sorted"] = [stats.sorted.name]
        data["Mitochondrial DNA type"] = [stats.mitochondrial_dna_model.name]
        data["Name type mtDNA"] = [stats.name_type_mtdna.name]
        data["Name type chromosomes"] = [stats.name_type_chromosomes.name]
        data["Gender"] = [stats.gender.name]
        data["File type"] = [stats.file_type.name]
        data["Path"] = [str(stats.path)]
        data["Size"] = [size]
        for key, value in label_map.items():
            if value not in data or data[value] is None:
                data[value] = [str(stats.__dict__[key])]
        data = TabularData(None, [TabularDataRow(x[0], x[1]) for x in data.items()])
        return data



import subprocess
from pathlib import Path

from helix.data.sequence_type import SequenceType
from helix.utility.external import External
from helix.naming.converter import Converter


class SequenceStatistics:
    def __init__(self, type, name, reference_length, mapped, unmapped) -> None:
        self.type: SequenceType = type
        self.name: str = name
        self.reference_length: int = reference_length
        self.unmapped: int = unmapped
        self.mapped: int = mapped


class IndexStatsCalculator:
    def __init__(self, file: Path, external=External()) -> None:
        if not file.exists():
            raise RuntimeError(f"Unable to find file {file.name}")

        self._file = file
        self._external = external

    def get_stats(self):
        stats: list[SequenceStatistics] = []

        stats_text = self._external.samtools(
            ["idxstats", self._file], stdout=subprocess.PIPE, wait=True
        )
        stats_text: list[str] = stats_text.decode().split("\n")
        for line in stats_text:
            # Line format is: name, length, # mapped, # unmapped
            # http://www.htslib.org/doc/samtools-idxstats.html
            chromosome_type = None
            line = line.strip()
            if line == "":
                continue
            elements = line.split("\t")
            name = elements[0]
            reference_length = int(elements[1])
            mapped = int(elements[2])
            unmapped = int(elements[3])

            chromosome_type = Converter.get_type(name)

            stats.append(
                SequenceStatistics(
                    chromosome_type, name, reference_length, mapped, unmapped
                )
            )

        return stats



from pathlib import Path

from helix.data.coverage_stats import CoverageStats
from helix.alignment_map.index_stats_calculator import SequenceStatistics
from helix.data.alignment_stats import AlignmentStats
from helix.data.chromosome_name_type import ChromosomeNameType
from helix.data.file_type import FileType
from helix.data.gender import Gender
from helix.data.mitochondrial_model_type import MitochondrialModelType
from helix.data.mitochondrial_name_type import MitochondrialNameType
from helix.data.sequence_type import SequenceType
from helix.data.sorting import Sorting
from helix.fasta.reference import Reference


class AlignmentMapFileInfo:
    def __init__(self) -> None:
        self.path: Path = None
        self.sorted: Sorting = None
        self.indexed: bool = None
        self.file_type: FileType = None
        self.reference_genome: Reference = None
        self.content: SequenceType = None
        self.mitochondrial_dna_model: MitochondrialModelType = None
        self.build: int = None
        self.name_type_chromosomes: ChromosomeNameType = None
        self.name_type_mtdna: MitochondrialNameType = None
        self.sequence_count: int = None
        self.alignment_stats: AlignmentStats = None
        self.index_stats: list[SequenceStatistics] = None
        self.coverage_stats: CoverageStats = None
        self.gender: Gender = None



import gzip
from math import sqrt
from pathlib import Path

from helix.utility.sequencers import Sequencers


class FASTQFile:
    def __init__(self, path: Path, sequencers: Sequencers = Sequencers()) -> None:
        self.path = path
        self._sequencers = sequencers

    def process_FASTQ(self, paired=True):
        line_count = 0
        char_count = 0
        sequencer_id = ""
        rcnt = 0
        rlen = 0
        rmean = 0
        rM2 = 0
        with gzip.open(self.path, "rt") as f:
            for line in f:
                if (
                    line_count % 4 == 2
                ):  # more common case; checking length of sequencer read
                    rlen += len(line)
                    if rlen > 1:
                        rcnt += 1
                        rdelta = rlen - rmean
                        rmean += rdelta / rcnt
                        rdelta2 = rlen - rmean
                        rM2 += rdelta * rdelta2
                elif (
                    line_count == 0
                ):  # Only at start; while skipping header or to read first sequncer ID
                    if (
                        line[0] == "#"
                    ):  # Skip header; do not count (new spec coming out with header)
                        continue
                    else:  # line[0] == '@':   # First line and sequencer label
                        sequencer = line.strip().split("\t")[0][1:]
                        sequencer_id = self._sequencers.determine_sequencer(sequencer)
                        line_count = 1
                elif (
                    line_count > 20000
                ):  # Only at end; Average first 5K segments (4 lines per segment)
                    line_count -= 1
                    break
                char_count += len(line)
                line_count += 1

        average_read_length = 0
        average_read_stddev = 0
        if rcnt > 2:
            rstd = sqrt(rM2 / (rcnt - 1))
            average_read_length = rmean
            average_read_stddev = rstd

        # Do a quick estimate based on the character count in the first 10,000 lines above
        # (2500 read segments) divided into the file size.
        segments = 0
        fastq_stats = self.path.stat()
        if fastq_stats.st_size > 0:
            segments = int(fastq_stats.st_size / (char_count / (line_count / 4)))
            segments *= 2 if paired else 1

        # e.g. return "Illumima", 660000000, 150
        return sequencer_id, segments, average_read_length
